from .app_store_objects import *


from fastapi import APIRouter, FastAPI, Request
from fastapi.staticfiles import StaticFiles
from starlette.responses import FileResponse 
from fastapi.middleware.cors import CORSMiddleware
import uvicorn
import os


router = APIRouter()


class Service_app_store:
    async def add_app(self, headers: dict[str, str], item: Add_App_Request) -> Add_App_Response:
        default_response = Add_App_Response()

        try:
            pass
        except Exception as e:
            print(f"Error: {e}")
            #default_response.error = str(e)
            #default_response.success = False

        return default_response

    async def search_app(self, headers: dict[str, str], item: Search_App_Request) -> Search_App_Response:
        default_response = Search_App_Response()

        try:
            pass
        except Exception as e:
            print(f"Error: {e}")
            #default_response.error = str(e)
            #default_response.success = False

        return default_response

    async def get_app_detail(self, headers: dict[str, str], item: Get_App_Detail_Request) -> Get_App_Detail_Response:
        default_response = Get_App_Detail_Response()

        try:
            pass
        except Exception as e:
            print(f"Error: {e}")
            #default_response.error = str(e)
            #default_response.success = False

        return default_response

    async def export_data(self, headers: dict[str, str], item: Export_Data_Request) -> Export_Data_Response:
        default_response = Export_Data_Response()

        try:
            pass
        except Exception as e:
            print(f"Error: {e}")
            #default_response.error = str(e)
            #default_response.success = False

        return default_response


def init(service_instance: Any):
    @router.post("/add_app/", tags=["app_store"])
    async def add_app(request: Request, item: Add_App_Request) -> Add_App_Response:
        item = Add_App_Request().from_dict(item.to_dict())
        headers = dict(request.headers.items())
        return (await service_instance.add_app(headers, item)).to_dict()

    @router.post("/search_app/", tags=["app_store"])
    async def search_app(request: Request, item: Search_App_Request) -> Search_App_Response:
        item = Search_App_Request().from_dict(item.to_dict())
        headers = dict(request.headers.items())
        return (await service_instance.search_app(headers, item)).to_dict()

    @router.post("/get_app_detail/", tags=["app_store"])
    async def get_app_detail(request: Request, item: Get_App_Detail_Request) -> Get_App_Detail_Response:
        item = Get_App_Detail_Request().from_dict(item.to_dict())
        headers = dict(request.headers.items())
        return (await service_instance.get_app_detail(headers, item)).to_dict()

    @router.post("/export_data/", tags=["app_store"])
    async def export_data(request: Request, item: Export_Data_Request) -> Export_Data_Response:
        item = Export_Data_Request().from_dict(item.to_dict())
        headers = dict(request.headers.items())
        return (await service_instance.export_data(headers, item)).to_dict()


def run(service_instance: Any, port: str, html_folder_path: str="", serve_html_under_which_url: str="/"):
    init(service_instance=service_instance)

    app = FastAPI()
    app.add_middleware(
        CORSMiddleware,
        allow_origins=['*'],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )
    app.include_router(
        router,
        prefix="/app_store",
    )

    if (html_folder_path != ""):
        if os.path.exists(html_folder_path) and os.path.isdir(html_folder_path):
            app.mount(serve_html_under_which_url, StaticFiles(directory=html_folder_path, html = True), name="web")
            @app.get(serve_html_under_which_url, response_model=str)
            async def index_page():
                return FileResponse(os.path.join(html_folder_path, 'index.html'))
            @app.exception_handler(404) #type: ignore
            async def custom_404_handler(_, __): #type: ignore
                return FileResponse(os.path.join(html_folder_path, 'index.html'))
            print(f"The website is running at: http://127.0.0.1:{port}/")
        else:
            print(f"Error: You should give me an absolute html_folder_path than {html_folder_path}")

    print(f"You can see the docs here: http://127.0.0.1:{port}/docs")
    uvicorn.run( #type: ignore
        app=app,
        host="0.0.0.0",
        port=int(port)
    ) 


if __name__ == "__main__":
    service_instance = Service_app_store()
    run(service_instance, port="6060")


import os
import platform
import time
import numpy as np
import cv2

from auto_everything.base import Terminal
t = Terminal(debug=True)


class CommonGUI():
    def __init__(self, project_name=""):
        if platform == "darwin":
            # OS X
            print("Sorry, the Mac m1,m2,m3,mx... is totally garbage in my eyes. \nIt even does not support VirtualBox and docker auto-start.\nTry to use Linux, for example, Ubuntu, my friend!")
            exit()
        try:
            import pyautogui as pyautogui
        except Exception:
            error = """
To use this module, you have to install pyautogui:
    sudo apt -y install python3-xlib
    sudo pip3 install pyautogui
    sudo apt -y install python3-tk python3-dev
    sudo apt -y install scrot
            """
            raise Exception(error)

        self.pyautogui = pyautogui
        self.pyautogui.FAILSAFE = False

        self._data_folder = project_name + "_data"
        self._init_data_structure()
        self._load_data()

        self._last_check = {
            "name": "yingshaoxo is a super man",
            "x": 0,
            "y": 0,
        }

    def _init_data_structure(self):
        if not t.exists(self._data_folder):
            os.mkdir(self._data_folder)

    def _load_data(self):
        paths = [os.path.join(self._data_folder, filename)
                 for filename in os.listdir(self._data_folder)]
        files = [path for path in paths if os.path.isfile(path)]
        files = [path for path in paths if os.path.basename(
            path).split('.')[-1] in ["png", "jpg"]]

        self.img_dict = {}
        for file in files:
            obj_name = '.'.join(os.path.basename(file).split('.')[:-1])
            self.img_dict.update({obj_name: file})

    def _make_sure_img_dict_exists(self):
        if self.img_dict == {}:
            print(f'You should put image files (png, jpg) with meaningful name into {self._data_folder} folder first!')
            print("You can do this with the function 'capture_screen_manually(picture_name=**)'")
            exit()

    def capture_screen_manually(self, picture_name):
        """
        screenshot with area selecting, based on gnome-screenshot
        """
        target_path = os.path.join(self._data_folder, picture_name + '.png')
        t.run_command(f'gnome-screenshot -a -f "{target_path}"')

    def delay(self, milliseconds):
        """
        delay for milliseconds
        """
        time.sleep(milliseconds / 1000)

    def find_text(self, target_text, from_image=None):
        """
        will return a list of center points

        a point: (x, y)
        """

        try:
            # from PIL import Image
            import pyscreenshot as ImageGrab
        except ImportError:
            error = """
To use this module, you have to install pyscreenshot:
    sudo pip3 install pyscreenshot
            """
            raise Exception(error)

        try:
            import pytesseract
        except ImportError:
            error = """
To use this module, you have to install tesseract:
    sudo apt -y install tesseract
    sudo pip3 -y install pytesseract
            """
            raise Exception(error)
        result = pytesseract.image_to_data(
            ImageGrab.grab(), output_type='dict')
        target_index = []
        for index, text in enumerate(result['text']):
            if target_text in text:
                target_index.append(index)
        return_list = []
        if (len(target_index) != 0):
            for t_index in target_index:
                list_ = ['top', 'left', 'width', 'height']
                my_dict = {}
                for attribute in list_:
                    value = result[attribute][t_index]
                    my_dict.update({attribute: value})
                x = my_dict['left'] + my_dict['width'] / 2
                y = my_dict['top'] + my_dict['height'] / 2
                return_list.append((x, y))
        if (len(return_list) != 0):
            return return_list
        else:
            return None


class GUI(CommonGUI):
    """
    A wrapper for pyautogui.
    """

    def __init__(self, project_name="", time_takes_for_one_click=0.2, grayscale=False):
        CommonGUI.__init__(self, project_name=project_name)
        self.__time_takes_for_one_click = time_takes_for_one_click
        self.__grayscale = grayscale

    def __click(self, x, y, game=False):
        self.pyautogui.click(x, y, interval=self.__time_takes_for_one_click)
        if game == True:
            self.pyautogui.click(x, y + 1, interval=self.__time_takes_for_one_click)

    def __get_tuple(self, element_name, confidence=0.9):
        self._make_sure_img_dict_exists()
        element_name = self.img_dict[element_name]
        return self.pyautogui.locateCenterOnScreen(element_name, confidence=confidence, grayscale=self.__grayscale)

    def hide_mouse(self):
        self.pyautogui.mouseUp()
        self.delay(1)
        self.pyautogui.moveTo(0, 0)

    def exists(self, element_name, confidence=0.9):
        """
        Check if an image exists at the screen

        Parameters
        ----------
        element_name: string
            image name (those pictures you put into data folder)
        """
        print(f"Ask for {element_name}")
        Tuple = self.__get_tuple(
            element_name, confidence=confidence,
        )
        if Tuple == None:
            return False
        else:
            self._last_check["name"] = element_name
            self._last_check["x"] = Tuple[0]
            self._last_check["y"] = Tuple[1]
            return True

    def get_center_point_of(self, element_name, confidence=0.9):
        if element_name == None or element_name == self._last_check["name"]:
            x = self._last_check["x"]
            y = self._last_check["y"]
        else:
            Tuple = self.__get_tuple(element_name, confidence=confidence)
            if Tuple != None:
                x, y = Tuple
            else:
                x, y = None, None
        return x, y

    def click(self, element_name=None, game=False):
        """
        click an image at the screen

        Parameters
        ----------
        element_name: string
            image name (those pictures you put into data folder) 
        game: bool
            for games, we click it in a different way
        """
        print(f"Try to click {element_name}")
        if element_name == None or element_name == self._last_check["name"]:
            x = self._last_check["x"]
            y = self._last_check["y"]
            self.__click(x, y, game=game)
        else:
            Tuple = self.__get_tuple(element_name)
            if Tuple != None:
                x, y = Tuple
                self.__click(x, y, game=game)


class Controller():
    """
    1. when you hit ctrl+caps , start to write command
    2. / means you want to search text on screen, generate ramdom number for different position, like vimum
    """

    def __init__(self):
        try:
            import pyautogui as pyautogui
        except Exception:
            error = """
To use this module, you have to install pyautogui:
    sudo apt -y install python3-xlib
    sudo pip3 install pyautogui
    sudo apt -y install python3-tk python3-dev
            """
            raise Exception(error)
        self.pyautogui = pyautogui

    def get_mouse_position(self):
        return self.pyautogui.position()

    def set_mouse_position(self, x, y):
        self.pyautogui.moveTo(x, y)

    def mouse_click(self, x=None, y=None, button='left', interval=0.5, game=False):
        self.pyautogui.click(x=x, y=y, button=button, interval=interval)
        if game == True:
            self.pyautogui.click(x=x + 1, y=y + 1, button=button, interval=interval)

    def write(self, text: str):
        self.pyautogui.write(text)

    def press(self, key: str):
        """
        self.pyautogui.KEYBOARD_KEYS

        or

        ['\t', '\n', '\r', ' ', '!', '"', '#', '$', '%', '&', "'", '(',
         ')', '*', '+', ',', '-', '.', '/', '0', '1', '2', '3', '4', '5', '6', '7',
         '8', '9', ':', ';', '<', '=', '>', '?', '@', '[', '\\', ']', '^', '_', '`',
         'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
         'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '{', '|', '}', '~',
         'accept', 'add', 'alt', 'altleft', 'altright', 'apps', 'backspace',
         'browserback', 'browserfavorites', 'browserforward', 'browserhome',
         'browserrefresh', 'browsersearch', 'browserstop', 'capslock', 'clear',
         'convert', 'ctrl', 'ctrlleft', 'ctrlright', 'decimal', 'del', 'delete',
         'divide', 'down', 'end', 'enter', 'esc', 'escape', 'execute', 'f1', 'f10',
         'f11', 'f12', 'f13', 'f14', 'f15', 'f16', 'f17', 'f18', 'f19', 'f2', 'f20',
         'f21', 'f22', 'f23', 'f24', 'f3', 'f4', 'f5', 'f6', 'f7', 'f8', 'f9',
         'final', 'fn', 'hanguel', 'hangul', 'hanja', 'help', 'home', 'insert', 'junja',
         'kana', 'kanji', 'launchapp1', 'launchapp2', 'launchmail',
         'launchmediaselect', 'left', 'modechange', 'multiply', 'nexttrack',
         'nonconvert', 'num0', 'num1', 'num2', 'num3', 'num4', 'num5', 'num6',
         'num7', 'num8', 'num9', 'numlock', 'pagedown', 'pageup', 'pause', 'pgdn',
         'pgup', 'playpause', 'prevtrack', 'print', 'printscreen', 'prntscrn',
         'prtsc', 'prtscr', 'return', 'right', 'scrolllock', 'select', 'separator',
         'shift', 'shiftleft', 'shiftright', 'sleep', 'space', 'stop', 'subtract', 'tab',
         'up', 'volumedown', 'volumemute', 'volumeup', 'win', 'winleft', 'winright', 'yen',
         'command', 'option', 'optionleft', 'optionright']
        """
        self.pyautogui.press(key)


class AndroidGUI(CommonGUI):
    """
    To interact with scrcpy by x11 and OpenCV.
    """

    def __init__(self, scrcpy_window_name, project_name=""):
        assert "not found" not in t.run_command("scrcpy -v"), """
To use this module, you have to install scrcpy
        """
        from myx11 import MyX11
        self.myx11 = MyX11()
        self.__window_name = scrcpy_window_name

        assert "not found" not in t.run_command("adb devices"), """
To use this module, you have to install adb:
    sudo apt install android-tools-adb
"""

        try:
            from uiautomator import device as d
            self.d = d
        except Exception as e:
            print(e)
            print(
                "I suggest you to use uiautomator to handle your android phone. You can install it by: \nsudo pip3 install uiautomator")

        info_of_android = t.run_command("adb shell dumpsys display")
        for line in info_of_android.split("\n"):
            if "StableDisplayWidth" in line:
                self.__android_width = int(line.split("=")[1].strip())
            if "StableDisplayHeight" in line:
                self.__android_height = int(line.split("=")[1].strip())

        self.__screen_height = None
        self.__screen_width = None

        CommonGUI.__init__(self, project_name=project_name)

    def capture_screen(self):
        """
        Get opencv image.
        """
        width, height, data = self.myx11.capture_screen(self.__window_name)
        data = cv2.cvtColor(np.array(data, np.uint8).reshape([height, width, 3]), cv2.COLOR_BGR2RGB)
        return data

    def find_all(self, element_name, from_image=None, threshold=0.8, visual=False):
        """
        find all images at the screen

        Parameters
        ----------
        element_name: string
            name for the sub image
        from_image: opencv image
        """
        self._make_sure_img_dict_exists()

        if not isinstance(from_image, np.ndarray):
            from_image = self.capture_screen()
        self.__screen_height = from_image.shape[0]
        self.__screen_width = from_image.shape[1]

        from_image_grayscale = cv2.cvtColor(from_image, cv2.COLOR_RGB2GRAY)
        template = cv2.imread(self.img_dict[element_name], cv2.IMREAD_GRAYSCALE)
        res = cv2.matchTemplate(from_image_grayscale, template, cv2.TM_CCOEFF_NORMED)

        loc = np.where(res >= threshold)
        result = []
        w, h = template.shape[::-1]
        for pt in zip(*loc[::-1]):
            x, y = (pt[0] + w / 2), (pt[1] + h / 2)
            result.append({"x": x, "y": y})
            if visual:
                cv2.rectangle(from_image, pt, (pt[0] + w, pt[1] + h), (0, 0, 255), 2)
        if visual:
            return from_image, result
        return result

    def find_one_automatically(self, element_name, from_image=None, threshold=0.5):
        times = 5
        while (times > 0 and 0 < threshold <= 1):
            times -= 1
            print(element_name)
            results = self.find_all(element_name=element_name, from_image=from_image, threshold=threshold)
            if len(results) == 1:
                return results[0]
            elif len(results) > 1:
                threshold += 0.1
            elif len(results) < 1:
                threshold -= 0.1
        return None

    def click(self, point, landscape=True):
        """
        do a click at android screen by using adb

        Parameters
        ----------
        point: {"x": 1920, "y": 1080}
            it's actually the result you get from function find_all().
        """
        assert "x" in point and "y" in point, f"You gave me a wrong point: {str(point)}"
        x = point["x"]
        y = point["y"]
        if landscape:
            x = int(x / self.__screen_width * self.__android_height)
            y = int(y / self.__screen_height * self.__android_width)
        else:
            x = int(x / self.__screen_width * self.__android_width)
            y = int(y / self.__screen_height * self.__android_height)
        t.run(f"adb shell input tap {x} {y}")


if __name__ == "__main__":
    androidGUI = AndroidGUI("pixel")
    while (True):
        data = androidGUI.capture_screen()
        cv2.imshow('frame', data)
        if cv2.waitKey(1) & 0xFF == ord('q'):
            break



from typing import Any

import random
import string
import hashlib
import base64
import json


class JWT_Tool():
    def my_jwt_encode(self, data: dict[str, Any], a_secret_string_for_integrity_verifying: str, use_md5: bool = True) -> str:
        """
        return the jwt string. it should only be used as a temporary password, you should not contain any sensitive information in it.

        Parameters
        ----------
        data: 
            a_dict, which contains the real information

        a_secret_string_for_integrity_verifying: string
            a secret string
        """
        header = {
            "alg": "SHA256",
            "typ": "JWT"
        }
        if (use_md5):
            header["alg"] = "MD5"
        header_base64_string = base64.b64encode(json.dumps(header).encode(encoding="utf-8")).decode("ascii")

        payload = data
        payload_base64_string = base64.b64encode(json.dumps(payload).encode(encoding="utf-8")).decode("ascii")

        if (use_md5):
            m = hashlib.md5()
        else:
            m = hashlib.sha256()

        m.update((header_base64_string + "." + payload_base64_string + "." + a_secret_string_for_integrity_verifying).encode(encoding="utf-8"))
        signature_ = m.hexdigest()

        return header_base64_string + "." + payload_base64_string + "." + signature_

    def my_jwt_decode(self, jwt_string: str, a_secret_string_for_integrity_verifying: str) -> dict[str, Any] | None:
        """
        return `None` if verifying didn't pass

        Parameters
        ----------
        jwt_string: 
            any string

        a_secret_string_for_integrity_verifying: string
            a secret string
        """
        try:
            splits = jwt_string.split(".")
            if len(splits) != 3:
                return None
            
            header = json.loads(base64.b64decode(splits[0]).decode(encoding="utf-8")) 
            header_base64_string = base64.b64encode(json.dumps(header).encode(encoding="utf-8")).decode("ascii")

            payload = json.loads(base64.b64decode(splits[1]).decode(encoding="utf-8"))
            payload_base64_string = base64.b64encode(json.dumps(payload).encode(encoding="utf-8")).decode("ascii")

            if (header['alg'] == "MD5"):
                m = hashlib.md5()
            else:
                m = hashlib.sha256()

            m.update((header_base64_string + "." + payload_base64_string + "." + a_secret_string_for_integrity_verifying).encode(encoding="utf-8"))
            signature_ = m.hexdigest()

            if signature_ == splits[2]:
                return payload
            else:
                return None
        except Exception as e:
            print(f"error: {e}")
            return None

    def yingshaoxo_json_web_token_encode(self, data: dict[str, Any], a_secret_string_for_integrity_verifying: str) -> str:
        """
        return the yingshaoxo's jwt string. it can has sensitive information.

        Parameters
        ----------
        data: 
            a_dict, which contains the real information

        a_secret_string_for_integrity_verifying: string
            a secret string
        """
        from auto_everything.disk import Disk
        encryption_and_decryption = Encryption_And_Decryption()
        disk = Disk()

        secret_dict = encryption_and_decryption.get_secret_alphabet_dict(a_secret_string=a_secret_string_for_integrity_verifying)

        json_string = json.dumps(data)
        bytes_json = json_string.encode(encoding='utf-8', errors="ignore")
        base64_json_string = disk.bytes_to_base64(bytes_data=bytes_json)
        encrypted_base64_string = encryption_and_decryption.encode_message(a_secret_dict=secret_dict, message=base64_json_string)
        
        return encrypted_base64_string

    def yingshaoxo_json_web_token_decode(self, jwt_string: str, a_secret_string_for_integrity_verifying: str) -> dict[str, Any] | None:
        """
        return `None` if verifying didn't pass

        Parameters
        ----------
        jwt_string: 
            any string

        a_secret_string_for_integrity_verifying: string
            a secret string
        """
        from auto_everything.disk import Disk
        encryption_and_decryption = Encryption_And_Decryption()
        disk = Disk()

        secret_dict = encryption_and_decryption.get_secret_alphabet_dict(a_secret_string=a_secret_string_for_integrity_verifying)

        try:
            decrypted_base64_string = encryption_and_decryption.decode_message(a_secret_dict=secret_dict, message=jwt_string)
            bytes_json = disk.base64_to_bytes(base64_string=decrypted_base64_string)
            json_string = bytes_json.decode(encoding="utf-8", errors="ignore")
            json_object = json.loads(json_string)
            return json_object
        except Exception as e:
            print(e)
            return None


class Password_Generator():
    """
    This is for hash password generation
    """

    def __init__(self, base_secret_string: str):
        """
        base_string: a base_string for PassWord generation
        """
        self.__base_data = base_secret_string.encode("utf-8")

    def get_password(self, secret_string_list: list[str], length: int = 12):
        """
        return a hashed string based on the text you were given
        you can use the returned string as your password

        Parameters
        ----------
        secret_string_list: 
            any string

        length: int
            password length
        """
        if (len(secret_string_list) == 0):
            raise Exception("You should give me at least one secret string in secret_string_list")

        data_list = [string.encode("utf-8") for string in secret_string_list]

        m = hashlib.sha512()
        m.update(self.__base_data)

        for data in data_list:
            m.update(data)

        result = m.hexdigest()
        result = result[:length - 1]
        # result = "A" + result
        return result

    def get_random_password(self, 
                            use_numbers: bool = True, 
                            use_ascii_lowercase_characters: bool = True, 
                            use_ascii_uppercase_characters: bool = True, 
                            use_punctuation: bool = False, 
                            additional_string_at_head: str = "",
                            length: int = 12):
        """
        return a random string, you can use the returned string as your password

        Parameters
        ----------
        additional_string_at_head: str
            You can put some string like '@yingshaoxo_' in this argument, so for each password you get, it will start with '@yingshaoxo_'

        length: int
            password length
        """
        characters_list: list[str] = []
        if (use_numbers == True):
            characters_list += list(string.digits)
        if (use_ascii_lowercase_characters == True):
            characters_list += list(string.ascii_lowercase)
        if (use_ascii_uppercase_characters == True):
            characters_list += list(string.ascii_uppercase)
        if (use_punctuation == True):
            characters_list += list(string.punctuation)
        result = ''.join(random.choices(characters_list, k=length))
        return additional_string_at_head + result


class Encryption_And_Decryption():
    def get_secret_alphabet_dict(self, a_secret_string: str) -> dict[str, str]:
        a_secret_string = a_secret_string.replace(" ", "").lower()
        character_list: list[str] = []
        for char in a_secret_string:
            if char.isalpha():
                if char not in character_list:
                    character_list.append(char)

        if len(character_list) >= 26:
            character_list = character_list[:26]
        else:
            characters_that_the_key_didnt_cover: list[str] = []
            for char in list(string.ascii_lowercase):
                if char not in character_list:
                    characters_that_the_key_didnt_cover.append(char)
            character_list = character_list + characters_that_the_key_didnt_cover

        final_dict: dict[str, str] = {}

        # for alphabet
        for index, char in enumerate(list(string.ascii_lowercase)):
            final_dict[char] = character_list[index]
        
        # for numbers
        original_numbers_in_alphabet_format = list(string.ascii_lowercase)[:10] #0-9 representations in alphabet format
        secret_numbers_in_alphabet_format = list(final_dict.values())[:10]
        final_number_list: list[str] = []
        for index in range(0, 10):
            secret_char = secret_numbers_in_alphabet_format[index]
            if secret_char in original_numbers_in_alphabet_format:
                final_number_list.append(str(original_numbers_in_alphabet_format.index(secret_char)))
        if len(final_number_list) >= 10:
            final_number_list = final_number_list[:10]
        else:
            numbers_that_didnt_get_cover: list[str] = []
            for char in range(0, 10):
                char = str(char)
                if char not in final_number_list:
                    numbers_that_didnt_get_cover.append(char)
            final_number_list = final_number_list + numbers_that_didnt_get_cover
        for index, char in enumerate(final_number_list):
            final_dict[str(index)] = char
        
        return final_dict

    def encode_message(self, a_secret_dict: dict[str, str], message: str) -> str:
        new_message: str = ""
        for char in message:
            if (not char.isalpha()) and (not char.isnumeric()):
                new_message += char
                continue
            if char in a_secret_dict.keys():
                new_char = a_secret_dict[char.lower()]
                if char.isupper():
                    new_char = new_char.upper()
            else:
                new_char = char
            new_message += new_char
        return new_message
    
    def decode_message(self, a_secret_dict: dict[str, str], message: str) -> str:
        new_secret_dict: dict[str, str] = {}
        for key, value in a_secret_dict.items():
            new_secret_dict[value] = key
        a_secret_dict = new_secret_dict

        new_message: str = ""
        for char in message:
            if (not char.isalpha()) and (not char.isnumeric()):
                new_message += char
                continue
            if char in a_secret_dict.keys():
                new_char = a_secret_dict[char.lower()]
                if char.isupper():
                    new_char = new_char.upper()
            else:
                new_char = char
            new_message += new_char
        return new_message
    
    def encrypt_a_code_folder(self, source_folder: str, target_folder: str, type_limiter: list[str] = ['.py', '.ts', '.js', '.vue', 'kt', '.java', '.go', '.rs', '.c', '.cpp'], password: str = "yingshaoxo is the best, no one can deny that!!!", text_handle_function: Any = None, gitignore_text: str|None = None):
        from auto_everything.io import IO   
        from auto_everything.disk import Disk
        io_ = IO()
        disk = Disk()

        real_code_folder = source_folder
        fake_code_folder = target_folder

        disk.delete_a_folder(fake_code_folder)
        disk.copy_a_folder(real_code_folder, fake_code_folder)

        password_dict = self.get_secret_alphabet_dict(password)

        files = disk.get_files(fake_code_folder, recursive=True, type_limiter=type_limiter, gitignore_text=gitignore_text)
        for file in files:
            text = io_.read(file)

            if (text_handle_function == None):
                new_text = self.encode_message(password_dict, text)
            else:
                new_text = text_handle_function(text)

            io_.write(file, new_text)

    def decrypt_a_code_folder(self, source_folder: str, target_folder: str, type_limiter: list[str] = ['.py', '.ts', '.js', '.vue', 'kt', '.java', '.go', '.rs', '.c', '.cpp'], password: str = "yingshaoxo is the best, no one can deny that!!!", text_handle_function: Any = None, gitignore_text: str|None = None):
        from auto_everything.io import IO   
        from auto_everything.disk import Disk
        io_ = IO()
        disk = Disk()

        fake_code_folder = source_folder
        predict_code_folder = target_folder

        disk.delete_a_folder(predict_code_folder)
        disk.copy_a_folder(fake_code_folder, predict_code_folder)

        password_dict = self.get_secret_alphabet_dict(password)

        files = disk.get_files(predict_code_folder, recursive=True, type_limiter=type_limiter, gitignore_text=gitignore_text)
        for file in files:
            text = io_.read(file)

            if (text_handle_function == None):
                new_text = self.decode_message(password_dict, text)
            else:
                new_text = text_handle_function(text)

            io_.write(file, new_text)

    def check_if_two_folder_equals(self, source_folder: str, target_folder, type_limiter: list[str] = ['.py', '.ts', '.js', '.vue', 'kt', '.java', '.go', '.rs', '.c', '.cpp']):
        from auto_everything.io import IO   
        from auto_everything.disk import Disk
        io_ = IO()
        disk = Disk()

        files1 = disk.get_files(source_folder, recursive=True, type_limiter=type_limiter)
        files1.sort()
        files2 = disk.get_files(target_folder, recursive=True, type_limiter=type_limiter)
        files2.sort()

        equal = True
        for index, file1 in enumerate(files2):
            file2 = files2[index]
            text1 = io_.read(file1)
            text2 = io_.read(file2)
            if (text1 != text2):
                equal = False
                print(f"Text do not eaual between files:")
                print(f"        * {file1}")
                print(f"        * {file2}")
                print()
                inputs = input(f"Ignore it and go to the next one? (y/n)").strip()
                if inputs.lower() != 'n':
                    continue
                else:
                    exit()
        
        
if __name__ == "__main__":
    # encryption_and_decryption = EncryptionAndDecryption()

    # a_dict = encryption_and_decryption.get_secret_alphabet_dict("hello, world")

    # a_sentence = "I'm yingshaoxo. Here is the test number: 9111108848."

    # encrypted_sentence = encryption_and_decryption.encode_message(a_secret_dict=a_dict, message=a_sentence)
    # print()
    # print(encrypted_sentence)

    # decrypted_sentence = encryption_and_decryption.decode_message(a_secret_dict=a_dict, message=encrypted_sentence)
    # print(decrypted_sentence)


    # password_generator = Password_Generator("yingshaoxo")
    # print(password_generator.get_random_password(use_punctuation=False, additional_string_at_head="@yingshaoxo_"))


    jwt_tool  = JWT_Tool()

    secret = "secret is a secret"

    a_jwt_string = jwt_tool.my_jwt_encode(data={"name": "yingshaoxo"}, a_secret_string_for_integrity_verifying=secret, use_md5=True)
    print(a_jwt_string)

    original_dict = jwt_tool.my_jwt_decode(jwt_string=a_jwt_string, a_secret_string_for_integrity_verifying=secret)
    print(original_dict)

    fake_jwt_string = "eyJhbGciOiAiU0hBMjU2IiwgInR5cCI6ICJKV1QifQ==.eyJuYW1lIjogInlpbmdzaGFveG8ifQ==.53c4df02e99e2ce3d3f8d230c799e9f0cbe9963e484b45efe171f38ebe58d690"
    original_dict = jwt_tool.my_jwt_decode(jwt_string=fake_jwt_string, a_secret_string_for_integrity_verifying=secret)
    print(original_dict)


import time
import threading
import re
from pprint import pprint
import copy
from inspect import signature
import os, tty, termios, sys, shlex
from typing import Any, Callable


class Python():
    """
    Python model was intended to simplify python development
    """

    def __init__(self):
        from auto_everything.io import IO
        from auto_everything.base import OS
        from auto_everything.terminal import Terminal
        from auto_everything.disk import Disk
        self._io = IO()
        self._os = OS()
        self._t = Terminal()
        self._disk = Disk()

    def check_if_a_variable_is_a_function(self, function: Any) -> bool:
        return isinstance(function, Callable)

    def list_python_packages(self):
        """
        Return a list of python packages which installed in your computer
        """
        return self._os.list_python_packages()

    def install_package(self, package_name: str):
        """
        Parameters
        ----------
        package_name: string
            the python package you want to install
        """
        self._os.install_python_package(package_name)

    def uninstall_package(self, package_name: str):
        """
        Parameters
        ----------
        package_name: string
            the python package you want to uninstall
        """
        self._os.uninstall_python_package(package_name)

    class loop():
        def __init__(self, interval: int | float=1, thread:bool=False):
            """
            interval: inverval in seconds
            new_thread: do you want to open a new thread? True/False
            """
            self.thread = thread
            self.interval = interval

        def __call__(self, func: Any):
            """
            func: a function which you want to run forever
            """

            def new_function(*args: Any, **kwargs: Any):
                def while_function():
                    while 1:
                        try:
                            func(*args, **kwargs)
                            time.sleep(self.interval)
                        except Exception as e:
                            print(e)

                if self.thread is False:
                    while_function()
                else:
                    threading.Thread(target=while_function).start()

            return new_function

    def help(self, object_: Any):
        """
        get help information about class or function
        """
        if callable(object_):
            arguments = str(signature(object_))
            print(object_.__name__ + arguments)

            doc = object_.__doc__
            if doc:
                print(doc, '\n')
        else:
            methods = dir(object_)
            private_methods: list[str] = []
            public_methods: list[str] = []
            for method in methods:
                if method[:1] == "_":
                    private_methods.append(method)
                else:
                    public_methods.append(method)
            print(private_methods, '\n')
            pprint(public_methods)
            """
            callable_list = inspect.getmembers(object_, predicate=inspect.ismethod) + inspect.getmembers(object_,
                                                                                                         predicate=inspect.isclass)
            callable_list = [(one, one[1].__name__ + str(signature(one[1]))) for one in callable_list]
            not_callable_list = dir(object_)
            for one in callable_list:
                if one[0] in not_callable_list:
                    not_callable_list.remove(one[0])
            not_callable_list = [(one, "") for one in not_callable_list]
            private_methods = []
            public_methods = []
            for one in not_callable_list + callable_list:
                if one[0][:1] == "_":
                    private_methods.append(one[0])
                else:
                    public_methods.append(one)

            print(private_methods, "\n")
            [print(one[0], one[1]) for one in public_methods]
            """

    def fire(self, class_name: Any):
        """
        fire is a function that will turn any Python class into a command line interface
        """
        self.fire2(class_instance=class_name)
        # from fire import Fire #type: ignore
        # Fire(class_name)

    def fire2(self, class_instance: Any, new_arguments: list[Any] = []):
        """
        fire2 is a function that come from ying_shao_xo's wild thinking which turn any Python class into a user friendly command line interface
        @yingshaoxo, baby
        """
        type_dict = {
            'int': int,
            'str': str,
            'bool': bool,
            'float': float
        }
        def get_argument_name(text: str):
            if ":" not in text:
                return text.split('=')[0].strip()
            else:
                return text.split(':')[0].strip()
        def get_type_string(text: str):
            if ":" not in text:
                return None
            text = text.split(':')[1].strip().split('=')[0].strip()
            splits = [one.strip() for one in text.split("|") if one.strip() != ""]
            for one in splits:
                if one in type_dict.keys():
                    return one
            return "str"
        def get_type_function(text: str):
            if ":" not in text:
                return None
            text = text.split(':')[1].strip().split('=')[0].strip()
            splits = [one.strip() for one in text.split("|") if one.strip() != ""]
            for one in splits:
                if one in type_dict.keys():
                    return type_dict[one]
            return str

        if not callable(class_instance):
            return

        class_instance2 = class_instance()

        if len(new_arguments) == 0:
            original_command_line_arguments = sys.argv
        else:
            original_command_line_arguments = new_arguments
        command_line_arguments = original_command_line_arguments[1:]
        my_method_and_propertys: dict[str, Any] = {}
        function_string_list = []

        for each_string in vars(class_instance).keys():
            if (not each_string.startswith("_")):
                one = class_instance.__dict__[each_string]
                if callable(one):
                    # it is a sub_function
                    arguments = str(signature(one))
                    #print(one.__name__, arguments)

                    function_string_list.append(each_string)
                    my_method_and_propertys[each_string] = {
                        'function_name': each_string,
                        'function_instance': getattr(class_instance2, each_string),
                        'arguments': {
                            get_argument_name(one2):
                                {
                                    'type_string': get_type_string(one2), 
                                    'type_function': get_type_function(one2),
                                }
                            for one2 in 
                            re.sub(
                                r"'(.*?)'", "''", 
                                re.sub(r'"(.*?)"', '""', 
                                       arguments[1:-1])
                            )
                            .split(', ')[1:]
                        },
                        'arguments_list': [
                            {
                                'argument_name': get_argument_name(one2),
                                'type_string': get_type_string(one2), 
                                'type_function': get_type_function(one2),
                            }
                            for one2 in 
                            re.sub(
                                r"'(.*?)'", "''", 
                                re.sub(r'"(.*?)"', '""', 
                                       arguments[1:-1])
                            )
                            .split(', ')[1:]
                        ],
                        'arguments_string': arguments
                    }
        # print(command_line_arguments)
        # print(my_method_and_propertys)

        if len(command_line_arguments) == 0:
            if len(my_method_and_propertys) == 0:
                print("APIs:\n")
                for function_name in function_string_list:
                    argument_part = ', '.join(my_method_and_propertys[function_name]['arguments_string'].split(', ')[1:])
                    if (argument_part.strip() == ""):
                        print((function_name + "\n    " + "()").strip())
                    else:
                        print((function_name + "\n    " + "(" + argument_part).strip())
            else:
                # the user do not know how to use this program, so make a shell for them
                def print_seperate_line():
                    print("\n" + '-'*9 + "\n")
                def print_functions_info(start_with: str = ""):
                    start_with = start_with.strip()
                    for function_name in function_string_list:
                        if start_with != "":
                            if (function_name.startswith(start_with)):
                                print(function_name)
                        else:
                            print(function_name)
                def print_argument_info(function_name: str):
                    if (function_name in my_method_and_propertys.keys()):
                        argument_part = ', '.join(my_method_and_propertys[function_name]['arguments_string'].split(', ')[1:])[:-1]
                        print(argument_part)
                    else:
                        print(f"No such function: {function_name}")
                def print_chars(text: str):
                    print(text, end="", flush=True)
                def clear_screen():
                    os.system("clear")
                def get_char_input() -> tuple[str, int]:
                    #https://www.physics.udel.edu/~watson/scen103/ascii.html
                    fd = sys.stdin.fileno()
                    old_settings = termios.tcgetattr(fd)
                    try:
                        tty.setraw(sys.stdin.fileno())
                        char = sys.stdin.read(1)
                    finally:
                        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
                    return char, ord(char)
                def get_last_word_of_string(text:str) -> tuple[str, int]:
                    splits = text.split(" ")
                    splits = [one for one in splits if one.strip() != ""]
                    if (text[-1].strip() == ""):
                        return "", len(splits)
                    if " " not in text:
                        return "", len(splits)
                    return splits[-1], len(splits)
                final_command_line = f"{original_command_line_arguments[0]} "
                while True:
                    clear_screen()
                    last_word, how_many_words = get_last_word_of_string(final_command_line)

                    if how_many_words == 2 and final_command_line.endswith(" "):
                        current_function_name = final_command_line.split(" ")[1].strip()
                        print_argument_info(current_function_name)
                    elif how_many_words > 2:
                        current_function_name = final_command_line.split(" ")[1].strip()
                        print_argument_info(current_function_name)
                    else:
                        print_functions_info(start_with=last_word)

                    print_seperate_line()
                    print_chars(final_command_line)
                    char, char_id = get_char_input()
                    if (char_id == 27):
                        # esc key
                        clear_screen()
                        break
                    elif char_id in [3]:
                        # ctrl + c key
                        clear_screen()
                        break
                    elif char_id == 127:
                        # delete key
                        final_command_line = final_command_line[:-1]
                    elif char_id == 9:
                        # tab key
                        # do complete
                        if last_word == "":
                            continue
                        new_word = ""
                        if (how_many_words == 0):
                            final_command_line = f"{original_command_line_arguments[0]} "
                            continue
                        elif (how_many_words == 1):
                            final_command_line = f"{original_command_line_arguments[0]} "
                            continue
                        elif (how_many_words == 2):
                            # complete function name
                            possible_new_words = []
                            for function_string in function_string_list:
                                if function_string.startswith(last_word):
                                    possible_new_words.append(function_string)
                            if len(possible_new_words) == 1:
                                new_word = possible_new_words[0]
                                # complete a whole word
                                final_command_line = final_command_line[:-len(last_word)]
                                final_command_line += new_word
                            elif len(possible_new_words) == 0:
                                new_word = ""
                                # complete nothing
                            else:
                                # complete common word
                                temp_possible_new_words = [word.lstrip(last_word) for word in possible_new_words]
                                temp_possible_new_words.sort(key=len)
                                the_longest_word = temp_possible_new_words[-1]
                                common_chars = ""
                                end_i = 0
                                while True:
                                    end_i += 1
                                    current_common_chars = the_longest_word[0:end_i]
                                    if all([word.startswith(current_common_chars) for word in temp_possible_new_words]):
                                        common_chars = current_common_chars
                                    if end_i > len(the_longest_word):
                                        break
                                final_command_line += common_chars
                        else:
                            # complete arguments name
                            current_function_name = final_command_line.split(" ")[1].strip()
                            correct_arguments = my_method_and_propertys[current_function_name]["arguments"]
                            for argument_string, value_dict in correct_arguments.items():
                                if argument_string.startswith(last_word):
                                    new_word = argument_string
                                    break
                            if new_word != "":
                                final_command_line = final_command_line[:-len(last_word)]
                                final_command_line += f"--{new_word}="
                    elif char_id == 10 or char_id == 13:
                        # enter key
                        clear_screen()
                        print(final_command_line)
                        print()
                        self.fire2(class_instance=class_instance, new_arguments=shlex.split(final_command_line.strip()))
                        exit()
                    else:
                        if (final_command_line[-1].strip() == "") and (char.strip() == ""):
                            continue
                        final_command_line += char

            return

        method_name = command_line_arguments[0]
        un_named_arguments = [one for one in command_line_arguments[1:] if not one.startswith('--')]
        named_arguments = [one for one in command_line_arguments[1:] if one.startswith('--')]
        if (method_name in function_string_list):
            one_method = my_method_and_propertys[method_name]
            method_instance = one_method['function_instance']
            right_arguments = one_method['arguments']
            right_arguments_list = one_method['arguments_list']

            custom_arguments = {}

            # for argument that does not have '--name=value', for example "Tools push 'message'""
            for index, one in enumerate(un_named_arguments):
                if index < len(right_arguments_list):
                    argument_name = right_arguments_list[index]['argument_name']
                    argument_value = one.strip()
                    argument_type = right_arguments_list[index]['type_function']

                    if (argument_type != None):
                        custom_arguments[argument_name] = argument_type(argument_value)
                    else:
                        # no type info
                        if str(argument_value).replace('.','',1).isdigit():
                            custom_arguments[argument_name] = float(str(argument_value))
                        else:
                            custom_arguments[argument_name] = str(argument_value)

            # for argument that does have '--name=value'
            for one in named_arguments:
                argument_name = one[2:].split("=")[0]
                argument_value = "=".join(one[2:].split("=")[1:])
                argument_type = right_arguments[argument_name]['type_function']
                if (argument_type != None):
                    custom_arguments[argument_name] = argument_type(argument_value)
                else:
                    # no type info
                    if str(argument_value).replace('.','',1).isdigit():
                        custom_arguments[argument_name] = float(str(argument_value))
                    else:
                        custom_arguments[argument_name] = str(argument_value)

            #print(f"{method_name} {' '.join(custom_arguments)}")
            method_instance(**custom_arguments)
            return

    def make_it_runnable(self, py_file_path: str|None=None):
        """
        make python file runnable

        after use this function, you can run the py_file by: ./your_py_script_name.py
        """
        if py_file_path is None or self._t.exists(py_file_path):
            py_file_path = os.path.join(
                self._t.current_dir, sys.argv[0].strip('./'))
        if os.path.exists(py_file_path):
            codes = self._io.read(py_file_path)
            expected_first_line = '#!/usr/bin/env {}'.format(self._t.py_executable)
            splits = codes.split('\n')
            found = False
            for line in splits:
                if line == expected_first_line:
                    found = True
                    break
            if found == False:
                codes = expected_first_line + '\n' + codes
                self._io.write(py_file_path, codes)
                self._t.run_command('chmod +x {}'.format(py_file_path))

            if not self._disk.executable(py_file_path):
                self._t.run_command('chmod +x {}'.format(py_file_path))

    def make_it_global_runnable(self, py_file_path: str| None=None, executable_name: str | None=None):
        """
        make python file global runnable

        after use this function, you can run the py_file at anywhere by: your_py_script_name.py
        """
        self.make_it_runnable(py_file_path)

        auto_everything_config_folder = "~/.auto_everything"
        bin_folder = os.path.expanduser(os.path.join(auto_everything_config_folder, "bin"))
        if not self._t.exists(bin_folder):
            self._t.run_command(f"mkdir -p {bin_folder}")

        if py_file_path is None or not self._t.exists(py_file_path):
            py_file_path = os.path.join(
                self._t.current_dir, sys.argv[0].strip('./'))

        is_the_first_running = False 
        runnable_path = None

        if os.path.exists(py_file_path):
            if executable_name == None:
                _, executable_name = os.path.split(py_file_path)
            runnable_path = os.path.join(bin_folder, executable_name)
            if not os.path.exists(runnable_path):
                is_the_first_running = True

            # remove links that the real file has been moved, or, remove links that match this py_file_path but with a different executable name
            files = os.listdir(bin_folder)
            [self._t.run(f"cd {bin_folder}; rm {file}") for file in files if
             not os.path.exists(os.path.join(bin_folder, file))]
            files = self._disk.get_files(bin_folder, recursive=False)
            [self._t.run(f"rm {file}") for file in files if
             os.path.realpath(file) == py_file_path and file != runnable_path]

            self._t.run_command(f"ln -s {py_file_path} {runnable_path}")

            bashrc_path = self._t.fix_path(f"~/.bashrc")
            bashrc_target_line = f'export PATH="$PATH:{bin_folder}"'
            bashrc = self._io.read(bashrc_path)
            if bashrc_target_line not in bashrc.split("\n"):
                bashrc = bashrc + "\n" + bashrc_target_line
                self._t.run_command(f"touch {bashrc_path}")
                self._io.write(bashrc_path, bashrc)
        
        if is_the_first_running and runnable_path:
            print(f"\n\n------------------\n\nYou could run \n\nsource ~/.bashrc\n\nto get started!")
            print(f"\n\n------------------\n\nYou could run \n\n{runnable_path.split('/')[-1]} -- --completion\n\nto get bash completion scripts")

    def print(self, data: Any, limit: int=20):
        """
        print `function help info` or print `dict` with length limit (So you could see the structure easily)
        """
        if callable(data):
            self.help(data)
        else:
            data = copy.deepcopy(data)

            def infinite_loop(the_data: Any) -> Any:
                the_type = type(the_data)
                if the_type == str:
                    return the_data[:limit] + "..."
                elif the_type == dict:
                    for key, value in the_data.items():
                        the_data[key] = infinite_loop(value)
                    return the_data
                elif the_type == list:
                    return [infinite_loop(element) for element in the_data]
                else:
                    return the_data

            data = infinite_loop(data)
            pprint(data)
    
    # def _python_code_preprocess(self, python_code: str):
    #     """
    #     { \n } => {}
    #     ( \n ) => ()
    #     \"\"\" \n \"\"\" => add 4 space before every line
    #     \'\'\' \n \'\'\' => add 4 space before every line
    #     > a remind not to touch the comments followed by a function top defination
    #     """
    #     pass
    
    def generate_documentation_for_a_python_project(self, python_project_folder_path: str, markdown_file_output_folder_path: str, only_generate_those_functions_that_has_docstring: bool=True):
        # code_block_match_rule = r"""(?P<code_block>(?:[ \t]*)(?P<code_head>(?:(?:(?:@(?:.*)\s+)*)*(?:(?:class)|(?:(?:async\s+)*def)))[ \t]*(?:\w+)\s*\((?:.*?)\)(?:[ \t]*->[ \t]*(?:(.*)*))?:)(?P<code_body>(?:\n(?:)(?:[ \t]+[^\n]*)|\n)+))"""
        head_information_regex_rule = r"""(?P<class_or_function_top_defination>(?: *@(?:.*?)\n+)* *(?:\s+(?P<is_class>class)|(?P<is_function>def|async +def)) +(?:(?:\n|.)*?):\n+)(?P<documentation>(?:(?:\s+[\"\']{3}(?:(?:\s|.)*?)[\"|\']{3}\n+)?(?:[ \t]*?\#(?:.*?)\n+)*)*)?(?P<class_or_function_propertys>(?(is_class)((?![ \t]+(?:def|class) )(?:(?:.*?): *(?:.*?) *= *(?:.*?)\n)*)|(?:)))?"""
        for file in self._disk.get_files(folder=python_project_folder_path, recursive=True, type_limiter=[".py"]):
            file_name = self._disk.get_file_name(file)
            if file_name.startswith("_"):
                continue

            raw_content = self._io.read(file)
            result_list = re.findall(pattern=head_information_regex_rule, string=raw_content)
            result_list = [
                {
                    'class_or_function_top_defination': one[0],
                    'is_class': one[1] == 'class',
                    'is_function': one[2] == 'def',
                    'documentation': one[3][0] if len(one[3]) == 1 else one[3],
                    'class_or_function_propertys': one[4]
                } 
                for one in result_list
            ]

            text = ""
            for item in result_list:
                class_or_function_top_defination = item["class_or_function_top_defination"]

                # function_name = class_or_function_top_defination.split(" ")[1]
                # if function_name.startswith("_"):
                #     continue

                documentation = item['documentation']
                # documentation = '\n'.join([one[4:] for one in documentation.split('\n')])
                is_class = item["is_class"]
                class_or_function_propertys = item["class_or_function_propertys"]

                heading_space_counting = 0
                for char in class_or_function_top_defination:
                    if char == " ":
                        heading_space_counting += 1
                    else:
                        break

                documentation = documentation.rstrip()
                if only_generate_those_functions_that_has_docstring == True:
                    if len(documentation.strip()) == 0:
                        continue
                class_or_function_propertys = class_or_function_propertys.rstrip() if is_class else ''

                if len(documentation.strip()) != 0 and len(class_or_function_propertys) != 0:
                    text += f"""
{class_or_function_top_defination.rstrip()}
{documentation.rstrip()}
{class_or_function_propertys}
{' ' * heading_space_counting + ' ' * 4}pass
                    """
                elif len(documentation.strip()) != 0 and len(class_or_function_propertys) == 0:
                    text += f"""
{class_or_function_top_defination.rstrip()}
{documentation.rstrip()}
{' ' * heading_space_counting + ' ' * 4}pass
                    """
                elif len(documentation.strip()) == 0 and len(class_or_function_propertys) != 0:
                    text += f"""
{class_or_function_top_defination.rstrip()}
{class_or_function_propertys}
{' ' * heading_space_counting + ' ' * 4}pass
                    """
                elif len(documentation.strip()) == 0 and len(class_or_function_propertys) == 0:
                    text += f"""
{class_or_function_top_defination.rstrip()}
{' ' * heading_space_counting + ' ' * 4}pass
                    """
            
            markdown_template = f"""
# {file_name}

```python
{text.strip()}
```
            """
            
            output_file_path = self._disk.join_paths(markdown_file_output_folder_path, file_name[:-len(".py")] + ".md") 
            self._io.write(file_path=output_file_path, content=markdown_template)

if __name__ == "__main__":
    py = Python()
    py.print({
        "hi": "dasssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssshgggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggggglddlahkdhkashgdkagda",
        "ok": {
            "fuck": "dhaslhdhagkdhaksghdashkgsadsagdsalhgsahd",
            "dhashd": "dsahhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhagdagdas",
            "dsadf": [{"sdf": "sadddddddddddddddddddddddddddddddddd"}, "asddddddddddddddddddddddddd"]
        }
    }, limit=25)



from .app_store_objects import *

from typing import Any, Callable
import json
from urllib import request


def _send_a_post(url: str, data: dict, headers: dict | None=None) -> str:
    a_request = request.Request(url, method="POST")
    a_request.add_header('Content-Type', 'application/json')
    if headers != None:
        for key, value in headers.items():
            a_request.add_header(key, value)
    data = json.dumps(data, indent=4)
    data = data.encode("utf-8", errors="ignore")
    r_ = request.urlopen(a_request, data=data)
    content = r_.read().decode("utf-8", errors="ignore")
    return content


class Client_app_store:
    def __init__(self, service_url: str, headers: dict[str, Any] | None = None, error_handle_function: Callable[[str], None] | None = None, special_error_key: str = "__yingshaoxo's_error__", interceptor_function: Callable[[dict], None] | None = None):
        if (service_url.endswith("/")):
            service_url = service_url[:-1]
        self._service_url = service_url 

        if headers == None:
            headers = {}
        self._header = headers

        def _default_error_handle_function(error_message: str):
            print(f"errors: {error_message}")
        if error_handle_function == None:
            error_handle_function = _default_error_handle_function
        self._error_handle_function = error_handle_function

        self._special_error_key = special_error_key

        def _default_interceptor_function(data: dict):
            pass
        if interceptor_function == None:
            interceptor_function = _default_interceptor_function
        self._interceptor_function = interceptor_function

    def _get_reponse_or_error_by_url_path_and_input(self, sub_url: str, input_dict: dict[str, Any]):
        the_url = f"{self._service_url}/app_store/{sub_url}/"
        try:
            response = _send_a_post(url=the_url, data=input_dict, headers=self._header)
            json_response = json.loads(response)
            self._interceptor_function(json_response)
            return json_response
        except Exception as e: 
            return {self._special_error_key: str(e)}

    def add_app(self, item: Add_App_Request, ignore_error: bool | None=None) -> Add_App_Response | None:
        result = self._get_reponse_or_error_by_url_path_and_input("add_app", item.to_dict())
        if self._special_error_key in result.keys():
            if ((ignore_error == None) or ((ignore_error != None) and (not ignore_error))):
                self._error_handle_function(result[self._special_error_key])
            return None
        else:
            return Add_App_Response().from_dict(result)

    def search_app(self, item: Search_App_Request, ignore_error: bool | None=None) -> Search_App_Response | None:
        result = self._get_reponse_or_error_by_url_path_and_input("search_app", item.to_dict())
        if self._special_error_key in result.keys():
            if ((ignore_error == None) or ((ignore_error != None) and (not ignore_error))):
                self._error_handle_function(result[self._special_error_key])
            return None
        else:
            return Search_App_Response().from_dict(result)

    def get_app_detail(self, item: Get_App_Detail_Request, ignore_error: bool | None=None) -> Get_App_Detail_Response | None:
        result = self._get_reponse_or_error_by_url_path_and_input("get_app_detail", item.to_dict())
        if self._special_error_key in result.keys():
            if ((ignore_error == None) or ((ignore_error != None) and (not ignore_error))):
                self._error_handle_function(result[self._special_error_key])
            return None
        else:
            return Get_App_Detail_Response().from_dict(result)

    def export_data(self, item: Export_Data_Request, ignore_error: bool | None=None) -> Export_Data_Response | None:
        result = self._get_reponse_or_error_by_url_path_and_input("export_data", item.to_dict())
        if self._special_error_key in result.keys():
            if ((ignore_error == None) or ((ignore_error != None) and (not ignore_error))):
                self._error_handle_function(result[self._special_error_key])
            return None
        else:
            return Export_Data_Response().from_dict(result)


from typing import Any

from datetime import datetime, timedelta
from time import sleep
from multiprocessing import Process


class Time:
    def __init__(self):
        self.default_datetime_string_format = "%Y-%m-%d %H:%M:%S"
        self.datetime = datetime
        self.timedelta = timedelta

    def get_current_timestamp_in_10_digits_format(self) -> int:
        return int(datetime.now().timestamp())

    def convert_timestamp_to_string(self, timestamp: int, format: str="%Y-%m-%d %H:%M:%S") -> str:
        return datetime.fromtimestamp(timestamp).strftime(format)
    
    def convert_string_to_timestamp(self, time_string: str, format: str="%Y-%m-%d %H:%M:%S") -> int:
        return int(datetime.strptime(time_string, format).timestamp())
    
    def get_timestamp_from_datetime_object(self, datetime_object: datetime) -> int:
        return int(datetime_object.timestamp())
    
    def get_datetime_object_from_timestamp(self, timestamp: int) -> datetime:
        return datetime.fromtimestamp(timestamp)
    
    def convert_datetime_object_to_string(self, datetime_object: datetime, format: str="%Y-%m-%d %H:%M:%S") -> str:
        return datetime_object.strftime(format)
    
    def run_a_function_at_a_certain_time(self, the_function: Any, the_time: str, format: str="%Y-%m-%d %H:%M:%S", wait=True) -> Process | None:
        def run_a_function_when_time_arrive():
            target_time = self.convert_datetime_object_to_string(
                    self.get_datetime_object_from_timestamp(
                        self.convert_string_to_timestamp(time_string=the_time, format=format)
                    ), format=format
                )
            while True:
                current_time = self.convert_datetime_object_to_string(self.datetime.now(), format=format)
                print(current_time, target_time)
                if (current_time == target_time):
                    a_process = Process(target=the_function)
                    a_process.start()
                    a_process.join()
                    return None
                sleep(0.1)

        if wait == True:
            run_a_function_when_time_arrive()
            return None
        else:
            new_process = Process(target=run_a_function_when_time_arrive)
            new_process.start()
            return new_process

    def run_a_function_after_x_seconds(self, the_function: Any, seconds: int | float, wait=True) -> Process | None:
        def run_a_function():
            sleep(seconds)
            a_process = Process(target=the_function)
            a_process.start()
            a_process.join()

        if wait == True:
            run_a_function()
            return None
        else:
            new_process = Process(target=run_a_function)
            new_process.start()
            return new_process
    
    def run_a_function_every_x_seconds(self, the_function: Any, seconds: int | float, wait=True) -> Process | None:
        def run_it():
            while True:
                the_function()
                sleep(seconds)

        if wait == True:
            run_it()
            return None
        else:
            new_process = Process(target=run_it)
            new_process.start()
            return new_process


from __future__ import annotations
import base64
from dataclasses import dataclass
import datetime
import pathlib
import tempfile
from typing import Any, Iterable, List, Tuple
from pathlib import Path
import os
import re
import json
import hashlib
import unicodedata
import string
from io import BytesIO
from fnmatch import fnmatch
import shutil
import random

import copy
import struct
import sys

from auto_everything.terminal import Terminal
t = Terminal(debug=True)


class Common:
    def __init__(self):
        self._auto_everything_config_folder = os.path.expanduser("~/.auto_everything")
        # print(self._auto_everything_config_folder)
        if not os.path.exists(self._auto_everything_config_folder):
            os.mkdir(self._auto_everything_config_folder)

    def _create_a_sub_config_folder_for_auto_everything(self, module_name: str):
        sub_folder_path = os.path.join(self._auto_everything_config_folder, module_name)
        if not os.path.exists(sub_folder_path):
            os.mkdir(sub_folder_path)
        return sub_folder_path


@dataclass
class _FileInfo:
    """Data Class for a single file returned by get_files()."""
    path: str
    is_folder: bool
    is_file: bool
    folder: str
    name: str
    level: int
    # parent: _FileInfo | None = None # You could try to make a global dict[path, _FilelInfo], then iterate twice to set parent
    children: List[_FileInfo] | None = None


class Sha256:
    # I recommend to use md5 or sha1, because this is slow for big file
    ks = [
        0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5,
        0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5,
        0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3,
        0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174,
        0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc,
        0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
        0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7,
        0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967,
        0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13,
        0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85,
        0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3,
        0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
        0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5,
        0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3,
        0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208,
        0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2,
    ]

    hs = [
        0x6a09e667, 0xbb67ae85, 0x3c6ef372, 0xa54ff53a,
        0x510e527f, 0x9b05688c, 0x1f83d9ab, 0x5be0cd19,
    ]

    M32 = 0xFFFFFFFF

    def __init__(self, m = None):
        self.mlen = 0
        self.buf = b''
        self.k = self.ks[:]
        self.h = self.hs[:]
        self.fin = False
        if m is not None:
            self.update(m)

    @staticmethod
    def pad(mlen):
        mdi = mlen & 0x3F
        length = (mlen << 3).to_bytes(8, 'big')
        padlen = 55 - mdi if mdi < 56 else 119 - mdi
        return b'\x80' + b'\x00' * padlen + length

    @staticmethod
    def ror(x, y):
        return ((x >> y) | (x << (32 - y))) & Sha256.M32

    @staticmethod
    def maj(x, y, z):
        return (x & y) ^ (x & z) ^ (y & z)

    @staticmethod
    def ch(x, y, z):
        return (x & y) ^ ((~x) & z)

    def compress(self, c):
        w = [0] * 64
        w[0 : 16] = [int.from_bytes(c[i : i + 4], 'big') for i in range(0, len(c), 4)]

        for i in range(16, 64):
            s0 = self.ror(w[i - 15],  7) ^ self.ror(w[i - 15], 18) ^ (w[i - 15] >>  3)
            s1 = self.ror(w[i -  2], 17) ^ self.ror(w[i -  2], 19) ^ (w[i -  2] >> 10)
            w[i] = (w[i - 16] + s0 + w[i - 7] + s1) & self.M32

        a, b, c, d, e, f, g, h = self.h

        for i in range(64):
            s0 = self.ror(a, 2) ^ self.ror(a, 13) ^ self.ror(a, 22)
            t2 = s0 + self.maj(a, b, c)
            s1 = self.ror(e, 6) ^ self.ror(e, 11) ^ self.ror(e, 25)
            t1 = h + s1 + self.ch(e, f, g) + self.k[i] + w[i]

            h = g
            g = f
            f = e
            e = (d + t1) & self.M32
            d = c
            c = b
            b = a
            a = (t1 + t2) & self.M32

        for i, (x, y) in enumerate(zip(self.h, [a, b, c, d, e, f, g, h])):
            self.h[i] = (x + y) & self.M32

    def update(self, m):
        if m is None or len(m) == 0:
            return

        assert not self.fin, 'Hash already finalized and can not be updated!'

        self.mlen += len(m)
        m = self.buf + m

        for i in range(0, len(m) // 64):
            self.compress(m[64 * i : 64 * (i + 1)])

        self.buf = m[len(m) - (len(m) % 64):]

    def digest(self):
        if not self.fin:
            self.update(self.pad(self.mlen))
            self.digest = b''.join(x.to_bytes(4, 'big') for x in self.h[:8])
            self.fin = True
        return self.digest

    def hexdigest(self):
        tab = '0123456789abcdef'
        return ''.join(tab[b >> 4] + tab[b & 0xF] for b in self.digest())


class Disk:
    """
    Some useful functions to handle your disk
    """

    def __init__(self):
        self.temp_dir: str = tempfile.gettempdir()

    def _expand_user(self, path: str | pathlib.PosixPath):
        # print(type(path))
        new_path: str = ""
        if type(path) == pathlib.PosixPath:
            new_path = str(path.as_posix()) #type: ignore
        else:
            new_path = path  # type: ignore
        if len(new_path) > 0:
            if new_path[0] == "~":
                new_path = os.path.expanduser(new_path)
        return new_path

    def exists(self, path: str) -> bool:
        """
        Check if a file or folder exist.

        Parameters
        ----------
        path: string
            the file path
        """
        path = self._expand_user(path)
        return Path(path).exists()

    def is_directory(self, path: str) -> bool:
        """
        Check if it is a folder.

        Parameters
        ----------
        path: string
            the file path
        """
        return os.path.isdir(path)

    def executable(self, path: str) -> bool:
        return os.access(path, os.X_OK)

    def concatenate_paths(self, *path: str) -> str:
        return os.path.join(*path) # type: ignore

    def join_paths(self, *path: str) -> str:
        return self.concatenate_paths(*path)

    def join_relative_paths(self, path1: str, path2: str) -> str:
        """
        Join path like: /aa/bb/cc + .././../d.txt
        This function will only care the second relative path
        """
        if path2.startswith("."):
            if path2.startswith("../"):
                return self.join_relative_paths('/'.join(path1.split('/')[:-1]) + "/", path2.replace("../", "", 1))
            elif path2.startswith("./"):
                return self.join_relative_paths(path1, path2.replace("./", "", 1))
            else:
                return self.join_paths(path1, path2)
        else:
            return self.join_paths(path1, path2)

    def get_current_working_directory(self) -> str:
        """
        Similar to bash script: `cwd`
        """
        return os.getcwd()

    def _parse_gitignore_text_to_list(self, gitignore_text: str) -> list[str]:
        ignore_pattern_list = [line for line in gitignore_text.strip().split("\n") if line.strip() != ""]
        new_ignore_pattern_list:list[str] = []
        for pattern in ignore_pattern_list:
            if pattern.startswith("#"):
                continue
            if pattern.endswith("/"):
                new_ignore_pattern_list.append(pattern.removesuffix("/"))
                new_ignore_pattern_list.append(pattern + "*")
            else:
                new_ignore_pattern_list.append(pattern)
        return new_ignore_pattern_list

    def _file_match_the_gitignore_rule_list(self, start_folder: str, file_path: str, ignore_pattern_list: list[str]):
        if not start_folder.endswith("/"):
            start_folder = start_folder + "/"
        else:
            start_folder = start_folder

        if file_path.startswith("./"):
            file_path = file_path[2:]

        # if file_path.endswith(".ipynb_checkpoints"):
        #     pass

        match = False
        for pattern in ignore_pattern_list:
            if fnmatch(file_path.removeprefix(start_folder), pattern.removeprefix("./")):
                match = True
                break
        return match

    def get_files(
        self,
        folder: str,
        recursive: bool = True,
        type_limiter: List[str] | None = None,
        gitignore_text: str|None = None,
        use_gitignore_file: bool = False
    ) -> List[str]:
        """
        Get files recursively under a folder.

        Parameters
        ----------
        folder: string
        recursive: bool
        type_limiter: List[str]
            a list used to do a type filter, like [".mp3", ".epub"]
        gitignore_text: str
            similar to git's .gitignore file, if any file matchs any rule, it won't be inside of the 'return file list'
        use_gitignore_file: bool
            if true, this function will not return any file/folder that matchs .gitignore file rules
        """
        folder = self._expand_user(folder)
        assert os.path.exists(folder), f"{folder} is not exist!"
        if recursive == True:
            files:list[str] = []
            for root, dirnames, filenames in os.walk(folder):
                for filename in filenames:
                    file:str = self.join_paths(root, filename)
                    if os.path.isfile(file):
                        if type_limiter:
                            p = Path(file)
                            if len(type_limiter) == 0 or (p.suffix in type_limiter):
                                files.append(file)
                        else:
                            files.append(file)

        else:
            if type_limiter:
                files = [
                    os.path.join(folder, f)
                    for f in os.listdir(folder)
                    if len(type_limiter) == 0
                    or
                    (os.path.isfile(os.path.join(folder, f)) and Path(os.path.join(folder, f)).suffix in type_limiter)
                ]
            else:
                files = [
                    os.path.join(folder, f)
                    for f in os.listdir(folder)
                    if os.path.isfile(os.path.join(folder, f))
                ]

        if gitignore_text != None:
            ignore_pattern_list = self._parse_gitignore_text_to_list(gitignore_text=gitignore_text)

            result_files:list[str] = []
            for file in files:
                if self._file_match_the_gitignore_rule_list(
                    start_folder=folder,
                    file_path=file,
                    ignore_pattern_list=ignore_pattern_list
                ) == False:
                    result_files.append(file)

            files = result_files

        if use_gitignore_file:
            if "version" not in t.run_command("git --version").lower():
                print("error: git needs to get installed for using 'use_gitignore_file' paramater.")
                return files
            ignored_files = []
            git_folder_list = []
            for file in files:
                if "/.git/" in file:
                    a_git_folder = file.split("/.git/")[0]
                    if a_git_folder in git_folder_list:
                        continue
                    else:
                        git_folder_list.append(a_git_folder)

                    result = t.run_command(f"git ls-files --other --directory", cwd=a_git_folder).strip()
                    if len(result) != 0:
                        if result.lower().startswith("fatal"):
                            continue
                        ignored_files += result.split("\n")
            ignored_files = list(set(ignored_files))

            new_files = []
            for file in files:
                if file not in ignored_files:
                    ok = True
                    for git_folder in git_folder_list:
                        if file.startswith(git_folder + "/.git/"):
                            ok = False
                            break
                    if ok:
                        new_files.append(file)
            files = new_files

        return files

    def get_folder_and_files(
        self,
        folder: str,
        recursive: bool = True,
        type_limiter: List[str] | None = None,
        gitignore_text: str|None = None
    ) -> Iterable[_FileInfo]:
        """
        Get files recursively under a folder.

        Parameters
        ----------
        folder: string
        recursive: bool
        type_limiter: List[str]
            a list used to do a type filter, like [".mp3", ".epub"]
        gitignore_text: str
            similar to git's .gitignore file, if any file matchs any rule, it won't be inside of the 'return file list'
        """
        folder = self._expand_user(folder)
        assert os.path.exists(folder), f"{folder} is not exist!"

        ignore_pattern_list = []
        if gitignore_text != None:
            ignore_pattern_list = self._parse_gitignore_text_to_list(gitignore_text=gitignore_text)

        for root, dirnames, filenames in os.walk(folder):
            level = root.replace(folder, '').count(os.sep)
            if level == 0:
                level = 1

            for dirname in dirnames:
                abs_folder_path = os.path.join(root, dirname)

                if gitignore_text != None:
                    if self._file_match_the_gitignore_rule_list(
                        start_folder=folder,
                        file_path=abs_folder_path,
                        ignore_pattern_list=ignore_pattern_list,
                    ):
                        continue

                yield _FileInfo(
                    level=level,
                    path=abs_folder_path,
                    is_folder=True,
                    is_file=False,
                    folder=self.get_directory_name(abs_folder_path),
                    name=dirname,
                )

            for filename in filenames:
                abs_folder_path = os.path.join(root, filename)

                if type_limiter != None:
                    should_remain = False
                    if len(type_limiter) != 0:
                        for suffix in type_limiter:
                            if abs_folder_path.endswith(suffix):
                                should_remain = True
                                break
                    else:
                        should_remain = True
                    if should_remain == False:
                        continue

                if gitignore_text != None:
                    if self._file_match_the_gitignore_rule_list(
                        start_folder=folder,
                        file_path=abs_folder_path,
                        ignore_pattern_list=ignore_pattern_list,
                    ):
                        continue

                if os.path.isfile(abs_folder_path):
                    yield _FileInfo(
                    level=level,
                    path=abs_folder_path,
                    is_folder=False,
                    is_file=True,
                    folder=self.get_directory_name(abs_folder_path),
                    name=filename,
                )

            if recursive == False:
                break

    def _super_sort_key_function(self, element: str) -> int:
        text = ""
        for char in element:
            if char.isdigit():
                text += char
            else:
                break
        if len(text) == 0:
            return 0
        return int(text[:10])

    def get_folder_and_files_tree(
        self,
        folder: str,
        reverse: bool = False,
        type_limiter: List[str] | None = None,
        gitignore_text: str|None = None,
    ) -> _FileInfo:
        """
        Get files recursively under a folder.

        Parameters
        ----------
        folder: string
        type_limiter: List[str]
            a list used to do a type filter, like [".mp3", ".epub"]
        gitignore_text: str
            similar to git's .gitignore file, if any file matchs any rule, it won't be inside of the 'return file list'
        """
        root = _FileInfo(
            path=folder,
            is_folder=True,
            is_file=False,
            folder=self.get_directory_name(folder),
            name=self.get_file_name(folder),
            level=0,
            children=None
        )

        ignore_pattern_list = []
        if gitignore_text != None:
            ignore_pattern_list = self._parse_gitignore_text_to_list(gitignore_text=gitignore_text)

        def dive(node: _FileInfo):
            folder = node.path

            if not os.path.isdir(folder):
                return

            items = os.listdir(folder)
            if len(items) == 0:
                return

            files_and_folders: list[_FileInfo] = []
            for filename in items:
                file_path = os.path.join(folder, filename)
                if (os.path.isdir(file_path)) or (type_limiter == None) or (Path(file_path).suffix in type_limiter):
                    # save
                    #absolute_file_path = os.path.abspath(file_path)

                    if gitignore_text != None:
                        if self._file_match_the_gitignore_rule_list(
                            start_folder=node.path,
                            file_path=file_path,
                            ignore_pattern_list=ignore_pattern_list,
                        ):
                            continue

                    new_node = _FileInfo(
                        path=file_path,
                        is_folder=os.path.isdir(file_path),
                        is_file=os.path.isfile(file_path),
                        folder=self.get_directory_name(file_path),
                        name=self.get_file_name(file_path),
                        level=node.level + 1,
                        children=None
                    )
                    dive(node=new_node)
                    files_and_folders.append(
                        new_node
                    )
                else:
                    # drop
                    continue
            files_and_folders.sort(key=lambda node_: self._super_sort_key_function(node_.name), reverse=reverse)
            node.children = files_and_folders

        dive(root)

        return root

    def sort_files_by_time(self, files: List[str], reverse: bool = False):
        files.sort(key=os.path.getmtime, reverse=reverse)
        return files

    def get_absolute_path(self, path: str) -> str:
        if path.startswith("~"):
            path = t.fix_path(path)

        return os.path.abspath(path=path)

    def get_stem_and_suffix_of_a_file(self, path: str) -> Tuple[str, str]:
        """
        /hi/you/abc.txt -> ('abc', '.txt')
        """
        p = Path(path)
        return p.stem, p.suffix

    def get_directory_path(self, path: str):
        """
        /hi/you/abc.txt -> /hi/you
        """
        path = self._expand_user(path)
        return os.path.dirname(path)

    def get_directory_name(self, path: str):
        """
        /hi/you/abc.txt -> you
        /hi/you -> you
        """
        path = self._expand_user(path)
        if self.exists(path):
            if os.path.isfile(path):
                path = os.path.dirname(path)
        else:
            raise Exception(f"Sorry, I don't know if '{path}' is a folder or file, because folder can also has '.' inside.")
        return os.path.basename(path)

    def get_parent_directory_name(self, path: str):
        """
        /hi/you/abc.txt -> you
        /hi/you -> hi
        """
        path = self._expand_user(path)
        path = os.path.dirname(path)
        return os.path.basename(path)

    def get_parent_directory_path(self, path: str):
        """
        /hi/you/abc.txt -> /hi/you
        /hi/you -> /hi
        """
        path = path.rstrip("/")
        path = self._expand_user(path)
        path = os.path.dirname(path)
        return path

    def get_file_name(self, path: str):
        """
        /hi/you/abc.txt -> abc.txt
        """
        return os.path.split(path)[-1]

    def get_hash_of_a_file(self, path: str) -> str:
        """
        calculate the blake2s hash string based on the bytes of a file.

        Parameters
        ----------
        path: string
            the file path
        """
        segment_size = 1024 * 1024 * 1
        path = self._expand_user(path)
        with open(path, "rb") as f:
            file_hash = hashlib.blake2s()
            while True:
                #data = f.read(8192)
                data = f.read(segment_size)
                if not data:
                    break
                file_hash.update(data)
        return file_hash.hexdigest()

    def get_hash_of_a_file_by_using_yingshaoxo_method(self, path: str) -> str:
        """
        get hash string based on the bytes of a file by using yingshaoxo method.

        Parameters
        ----------
        path: string
            the file path
        """
        segment_size = 1024 * 1024 * 1
        path = self._expand_user(path)
        file_hash = 0
        with open(path, "rb") as f:
            operator_flag = True
            while True:
                #data = f.read(8192)
                data = f.read(segment_size)
                if not data:
                    break
                for one_byte in data:
                    if operator_flag == True:
                        file_hash += one_byte
                        operator_flag = False
                    else:
                        file_hash -= one_byte
                        operator_flag = True
        file_hash = file_hash % 100000000
        file_hash = str(file_hash)
        return file_hash

    def get_hash_of_a_file_by_using_sha1(self, path: str) -> str:
        """
        get the sha1 hash string based on the bytes of a file.

        Parameters
        ----------
        path: string
            the file path
        """
        segment_size = 1024 * 1024 * 1
        path = self._expand_user(path)
        with open(path, "rb") as f:
            file_hash = hashlib.sha1()
            while True:
                #data = f.read(8192)
                data = f.read(segment_size)
                if not data:
                    break
                file_hash.update(data)
        return file_hash.hexdigest()

    def get_hash_of_a_folder(self, folder_path: str, print_log: bool = False) -> str:
        """
        get the sha1 hash string for a folder.

        Parameters
        ----------
        folder_path: string
            the folder path
        """
        folder_path = self._expand_user(folder_path)
        raw_files = self.get_files(folder_path)

        folder_list = []
        file_list = []
        for file in raw_files:
            if self.is_directory(file):
                folder_list.append(file)
            else:
                file_list.append(file)

        folder_list.sort()
        file_list.sort()

        general_hash = hashlib.sha1()

        for file in file_list:
            if print_log:
                print(file)
            general_hash.update(self.get_hash_of_a_file_by_using_sha1(file).encode("utf-8", errors="ignore"))
        for folder in folder_list:
            general_hash.update(folder.encode("utf-8", errors="ignore"))

        return general_hash.hexdigest()

    def get_hash_of_a_path(self, path: str) -> str:
        """
        calculate the blake2s hash string based on path name.

        Parameters
        ----------
        path: string
            actually it can be any string
        """
        file_hash = hashlib.blake2s()
        file_hash.update(path.encode(encoding="UTF-8"))
        return file_hash.hexdigest()

    def get_safe_name(self, filename: str, replace_chars: str = " ") -> str:
        """
        get a valid file name by doing a replacement. (English only)

        Parameters
        ----------
        filename: string
            the unsafe filename
        replace_chars: string
            chars in replace_chars will be replaced by '_'.
        """
        valid_filename_chars = "-_.() %s%s" % (string.ascii_letters, string.digits)
        whitelist = valid_filename_chars
        char_limit = 255

        # replace spaces
        for r in replace_chars:
            filename = filename.replace(r, "_")

        # keep only valid ascii chars
        cleaned_filename = (
            unicodedata.normalize("NFKD", filename).encode("ASCII", "ignore").decode()
        )

        # keep only whitelisted chars
        cleaned_filename = "".join(c for c in cleaned_filename if c in whitelist)
        if len(cleaned_filename) > char_limit:
            print(
                "Warning, filename truncated because it was over {}. Filenames may no longer be unique".format(
                    char_limit
                )
            )
        return cleaned_filename[:char_limit]

    def get_file_size(
        self, path: str | None, level: str = "B", bytes_size: int | None = None
    ) -> int | None:
        """
        Get file size in the unit of  B, KB, MB.
        Parameters
        ----------
        path: string
            the file path
        level: string
            B, KB, or MB
        bytes_size: int
            a number represent the file size in bytes level
        """
        if bytes_size is None:
            if path != None:
                path = self._expand_user(path)
                file = Path(path)
                assert file.exists(), f"{path} is not exist!"
                bytes_size = file.stat().st_size
            else:
                raise Exception("You should give me a file_path or bytes_size")
        if level == "B":
            return int("{:.0f}".format(bytes_size))
        elif level == "KB":
            return int("{:.0f}".format(bytes_size / float(1 << 10)))
        elif level == "MB":
            return int("{:.0f}".format(bytes_size / float(1 << 20)))
        else:
            return bytes_size

    def get_folder_size(self, path: str, level: str = "B") -> int:
        """
        Get folder size in the unit of  B, KB, MB.
        Parameters
        ----------
        path: string
            the folder path
        level: string
            B, KB, or MB
        """
        folder_path = path
        total_size = 0

        for root, directories, files in os.walk(folder_path):
            for file in files:
                file_path = os.path.join(root, file)
                file_size = os.stat(file_path).st_size
                total_size += file_size

            for directory in directories:
                directory_path = os.path.join(root, directory)
                directory_size = os.stat(directory_path).st_size
                total_size += directory_size

        if level == "B":
            return int("{:.0f}".format(total_size))
        elif level == "KB":
            return int("{:.0f}".format(total_size / float(1 << 10)))
        elif level == "MB":
            return int("{:.0f}".format(total_size / float(1 << 20)))
        else:
            return total_size

    def compress(self, input_folder_path: str, output_zip_path: str, file_format: str = "zip") -> str:
        """
        compress files to a target.

        Parameters
        ----------
        input_folder_path: string
            the input folder for compressing
        output_zip_path: string
            the compressed output file path, like *.zip
        file_format: string
            “zip”, “tar”, “gztar”, “bztar”, or “xztar”
        """
        input_folder_path = self._expand_user(input_folder_path)
        output_zip_path = self._expand_user(output_zip_path)

        if not self.exists(input_folder_path):
            raise Exception(f"The input_folder_path '{input_folder_path}' should exists.")
        if not self.is_directory(input_folder_path):
            raise Exception(f"The input_folder_path '{input_folder_path}' should be an folder.")

        output_folder = self.get_directory_path(output_zip_path)
        self.create_a_folder(folder_path=output_folder)

        pure_output_zip_base_name, suffix = self.get_stem_and_suffix_of_a_file(output_zip_path)
        shutil.make_archive(
            root_dir=input_folder_path,
            base_dir='./',
            base_name=self.join_paths(output_folder, pure_output_zip_base_name),
            format=file_format
        )
        # t.run(f"zip -r -D {target} {' '.join(paths)}")

        return output_zip_path

    def uncompress(self, compressed_file_path: str, extract_folder_path: str, file_format: str = "zip") -> bool:
        """
        uncompress a file.

        Parameters
        ----------
        compressed_file_path: string
            the compressed file path
        extract_folder_path: string
            a folder where you want to put the uncompressed files into
        file_format: string
            “zip”, “tar”, “gztar”, “bztar”, or “xztar”
        """
        try:
            compressed_file_path = self._expand_user(compressed_file_path)
            extract_folder_path = self._expand_user(extract_folder_path)

            if not self.exists(compressed_file_path):
                raise Exception(f"The compressed_file_path '{compressed_file_path}' should exists.")
            if not self.exists(extract_folder_path):
                self.create_a_folder(folder_path=extract_folder_path)

            shutil.unpack_archive(filename=compressed_file_path, extract_dir=extract_folder_path, format=file_format)

            return True
        except Exception as e:
            print(f"error: {e}")
            return False
        # path = self._expand_user(path)
        # folder = self._expand_user(folder)
        # assert self.exists(path), f"{path} was not exist"

        # t.run(f"rm {folder} -fr")
        # t.run(f"mkdir -p {folder}")
        # assert self.exists(folder), f"{folder} was not exit"
        # try:
        #     suffix = Path(path).suffix
        #     if suffix == ".zip":
        #         t.run(f"unzip '{path}' -d '{folder}'")
        #         if len(os.listdir(folder)) == 1:
        #             t.run(f"cd '{folder}' && cd * && mv * .. -f")
        #     elif suffix == ".gz":
        #         t.run(f"tar zxfv '{path}' --directory '{folder}' --strip-components=1")
        #         if len(os.listdir(folder)) == 0:
        #             t.run(f"tar zxfv '{path}' --directory '{folder}'")
        #     if len(os.listdir(folder)):
        #         return True
        #     else:
        #         return False
        # except Exception as e:
        #     raise e

    def get_the_temp_dir(self) -> str:
        """
        Get system level temporary folder. It's normally `/tmp/`
        """
        return self.temp_dir

    def get_a_temp_folder_path(self):
        """
        We'll add a random hash_string after the temp directory, so you can get a path like '/tmp/xxssddf'
        """
        m = hashlib.sha256()
        m.update(str(datetime.datetime.now()).encode("utf-8"))
        m.update(''.join(random.choices(string.ascii_uppercase, k=10)).encode("utf-8"))
        temp_folder_path = os.path.join(self.temp_dir, m.hexdigest()[:27])
        return temp_folder_path

    def get_a_temp_file_path(self, filename: str):
        """
        We'll add a hash_string before the filename, so you can use this file path without any worry
        """
        m = hashlib.sha256()
        m.update(str(datetime.datetime.now()).encode("utf-8"))
        m.update(filename.encode("utf-8"))
        stem, suffix = self.get_stem_and_suffix_of_a_file(filename)
        tempFilePath = os.path.join(self.temp_dir, m.hexdigest()[:10] + suffix)
        return tempFilePath

    def create_a_new_folder_under_home(self, folder_name: str):
        folder_path = self._expand_user(f"~/{folder_name}")
        if not os.path.exists(folder_path):
            # os.mkdir(folder_path)
            t.run_command(f"mkdir -p {folder_path}")
        return folder_path

    def get_bytesio_from_a_file(self, filepath: str) -> BytesIO:
        with open(filepath, "rb") as fh:
            buffer = BytesIO(fh.read())
        return buffer

    def get_part_of_a_file_in_bytesio_format_from_a_file(self, file_path: str, file_segment_size_in_bytes: int, segment_number: int) -> BytesIO:
        if segment_number <= 0:
            raise Exception("segment_number should be > 0, for example, 1")
        with open(file_path, "rb") as fh:
            for _ in range(segment_number-1):
                fh.read(file_segment_size_in_bytes)
            buffer = BytesIO(fh.read(file_segment_size_in_bytes))
        return buffer

    def save_bytesio_to_file(self, bytes_io: BytesIO, file_path: str):
        bytes_io.seek(0)
        with open(file_path, "wb") as f:
            f.write(bytes_io.read())

    def base64_to_bytesio(self, base64_string: str):
        splits = base64_string.split(",")
        if len(splits) == 2:
            base64_string = splits[1]
        img_data = base64.b64decode(base64_string)
        return BytesIO(img_data)

    def bytesio_to_base64(self, bytes_io: BytesIO):
        bytes_io.seek(0)
        return base64.b64encode(bytes_io.getvalue()).decode()

    def bytes_to_base64(self, bytes_data: bytes):
        return base64.b64encode(bytes_data).decode()

    def base64_to_bytes(self, base64_string: str):
        splits = base64_string.split(",")
        if len(splits) == 2:
            base64_string = splits[1]
        img_data = base64.b64decode(base64_string)
        return img_data

    def hex_to_bytes(self, hex_string: str):
        return bytes.fromhex(hex_string)

    def bytes_to_hex(self, bytes_data: bytes):
        return bytes_data.hex()

    def remove_a_file(self, file_path: str):
        file_path = self._expand_user(file_path)
        if self.exists(file_path):
            os.remove(file_path)

    def delete_a_file(self, file_path: str):
        self.remove_a_file(file_path=file_path)

    def move_a_file(self, source_file_path: str, target_file_path: str):
        source_file_path = self._expand_user(source_file_path)
        target_file_path = self._expand_user(target_file_path)
        if source_file_path == target_file_path:
            return
        if self.exists(target_file_path):
            os.remove(target_file_path)
        os.rename(source_file_path, target_file_path)

    def move_a_folder(self, source_folder_path: str, target_folder_path: str):
        source_folder_path = self._expand_user(source_folder_path)
        target_folder_path = self._expand_user(target_folder_path)
        if source_folder_path == target_folder_path:
            return
        if self.exists(target_folder_path):
            self.delete_a_folder(target_folder_path)
        os.rename(source_folder_path, target_folder_path)

    def copy_a_file(self, source_file_path: str, target_file_path: str):
        source_file_path = self._expand_user(source_file_path)
        target_file_path = self._expand_user(target_file_path)
        if source_file_path == target_file_path:
            return
        self.create_a_folder(self.get_directory_path(target_file_path))
        shutil.copyfile(source_file_path, target_file_path)

    def convert_bytes_to_bytesio(self, bytes_data: bytes) -> BytesIO:
        bytes_io = BytesIO()
        bytes_io.write(bytes_data)
        bytes_io.seek(0)
        return bytes_io

    def create_a_folder(self, folder_path: str):
        folder_path = self._expand_user(folder_path)
        Path(folder_path).mkdir(parents=True, exist_ok=True)

    def copy_a_folder(self, source_folder_path: str, target_folder_path: str):
        source_folder_path = self._expand_user(source_folder_path)
        target_folder_path = self._expand_user(target_folder_path)

        if (not self.exists(target_folder_path)):
            self.create_a_folder(target_folder_path)
        else:
            if not self.is_directory(target_folder_path):
                self.delete_a_file(target_folder_path)

        try:
            shutil.copytree(source_folder_path, target_folder_path, dirs_exist_ok=True)
        except OSError as exc: # python >2.5
            try:
                shutil.copy(source_folder_path, target_folder_path)
            except Exception as e:
                print(e)

    def delete_a_folder(self, folder_path: str):
        folder_path = self._expand_user(folder_path)
        if (self.exists(folder_path)):
            shutil.rmtree(path=folder_path)

    def fake_folder_backup(self, backup_folder: str, backup_saving_file_path: str | None=None) -> list[Any]:
        saving_path = None
        if backup_saving_file_path != None:
            saving_path = backup_saving_file_path
        files = self.get_folder_and_files(folder=backup_folder)
        data_list: list[Any] = []
        for file_or_folder in files:
            data_list.append({
                "path": file_or_folder.path,
                "type": 'folder' if os.path.isdir(file_or_folder.path) else 'file'
            })
        if (saving_path != None):
            with open(saving_path, 'w', encoding="utf-8") as f:
                f.write(json.dumps(data_list, indent=4))
            print(f"fake backup is done, it is in: {saving_path}")
        return data_list

    def fake_folder_recover(self, backup_saving_file_path: str):
        if not disk.exists(backup_saving_file_path):
            raise Exception(f"file does not exists: {backup_saving_file_path}")
        with open(backup_saving_file_path, 'r', encoding='utf-8') as f:
            raw_json = f.read()
            json_object = json.loads(raw_json)
        for item in json_object:
            path = item['path']
            type = item['type']
            if type == 'folder':
                os.mkdir(path)
                print(path)
            else:
                with open(path, 'w', encoding='utf-8') as f:
                    f.write("")
                print(path)
        print("\nfake recover is done, sir.")

    def read_bytes_from_file(self, file_path: str) -> bytes:
        """
        read bytes from a file

        Parameters
        ----------
        file_path: str
            like `/home/yingshaoxo/file.sql`
        """
        with open(file_path, 'rb') as f:
            result = f.read()
        return result

    def write_bytes_into_file(self, file_path: str, content: bytes):
        """
        write bytes into a file

        Parameters
        ----------
        file_path: str
            target file path
        content: bytes
        """
        with open(file_path, 'wb') as f:
            f.write(content)

    def convert_file_suffix_end_to_lowercase(self, source_folder: str):
        files = list(disk.get_files(source_folder, recursive=True))
        for file in files:
            file = disk.get_absolute_path(file)
            if ("/" in file):
                last_part = file.split("/")[-1]
                if ("." in last_part):
                    end = last_part.split(".")[-1]
                    if file.endswith("." + end):
                        new_file = file[:-len(end)] + end.lower()
                        disk.move_a_file(source_file_path=file, target_file_path=new_file)


class Store:
    """
    A key-value store.
    """

    def __init__(self, store_name: str, save_to_folder: str | None = None):
        self._common = Common()

        if save_to_folder is None or not os.path.exists(save_to_folder):
            self._store_folder = (
                self._common._create_a_sub_config_folder_for_auto_everything("store")
            )
        else:
            self._store_folder = save_to_folder

        self._store_name = store_name.strip()
        self.__initialize_SQL()

    def __initialize_SQL(self):
        import sqlite3 as sqlite3

        self._SQL_DATA_FILE = os.path.join(self._store_folder, f"{self._store_name}.db")
        self._sql_conn = sqlite3.connect(self._SQL_DATA_FILE, check_same_thread=False)

        def regular_expression(expr: str, item: Any):
            reg = re.compile(expr, flags=re.DOTALL)
            return reg.search(item) is not None

        self._sql_conn.create_function(
            "REGEXP", 2, regular_expression
        )  # 2 here means two parameters. REGEXP is a fixed value

        self._sql_cursor = self._sql_conn.cursor()
        self._sql_cursor.execute(
            f"""CREATE TABLE IF NOT EXISTS {self._store_name}
                    (key TEXT, value TEXT)"""
        )

    def __pre_process_key(self, key: Any):
        key = str(key)
        return key

    def __pre_process_value(self, value: Any):
        if not isinstance(value, str):
            try:
                value = json.dumps(value)
            except Exception as e:
                print(e)
                raise Exception(
                    f"The value you gave me is not a json object: {str(value)}"
                )
        return value

    def __active_json_value(self, value: Any):
        try:
            value = json.loads(value)
        except Exception as e:
            value = value
        return value

    def get_items(self):
        """
        get all key and value tuple in the store
        """
        rows:list[Any] = []
        for row in self._sql_cursor.execute(
            f"SELECT * FROM {self._store_name} ORDER BY key"
        ):
            rows.append((row[0], self.__active_json_value(row[1])))
        return rows

    def has_key(self, key: str) -> bool:
        """
        check if a key exist in the store

        Parameters
        ----------
        key: string
        """
        key = self.__pre_process_key(key)

        results = self._sql_cursor.execute(
            f'SELECT EXISTS(SELECT 1 FROM {self._store_name} WHERE key="{key}" LIMIT 1)'
        )
        if self._sql_cursor.fetchone()[0] > 0:
            return True
        else:
            return False

    def get(self, key: str, default_value: Any = None):
        """
        get a value by using a key

        Parameters
        ----------
        key: string
        default_value: string or an object that jsonable
        """
        key = self.__pre_process_key(key)

        self._sql_cursor.execute(
            f"SELECT * FROM {self._store_name} WHERE key=?", (key,)
        )
        result = self._sql_cursor.fetchone()
        if result:
            return self.__active_json_value(result[1])
        else:
            return default_value

    def set(self, key: str, value: Any):
        """
        set a value by using a key

        Parameters
        ----------
        key: string
        value: string or an object that jsonable
        """
        key = self.__pre_process_key(key)
        value = self.__pre_process_value(value)

        if self.has_key(key):
            self._sql_cursor.execute(
                f"UPDATE {self._store_name} SET value=? WHERE key=?", (value, key)
            )
        else:
            command = f""" INSERT INTO {self._store_name} VALUES(?,?) """
            self._sql_cursor.execute(command, (key, value))
        self._sql_conn.commit()

    def delete(self, key: str):
        """
        delete a value by using a key

        Parameters
        ----------
        key: string
        """
        key = self.__pre_process_key(key)

        if self.has_key(key):
            self._sql_cursor.execute(
                f"DELETE FROM {self._store_name} WHERE key=?", (key,)
            )

        self._sql_conn.commit()

    def reset(self):
        """
        empty the store
        """
        self._sql_cursor.execute(f"DELETE FROM {self._store_name}")
        self._sql_conn.commit()


class Dart_File_Hard_Encoder_And_Decoder:
    """
    Put a folder into your code as a dart file (yingshaoxo)
    """
    def __init__(self, generated_file_name="built_in_files.dart"):
        from auto_everything.io import IO
        from auto_everything.disk import Disk
        import json
        self._disk = Disk()
        self._io = IO()
        self._json = json

    def _get_content_json_string(self, source_folder: str) -> str:
        files = self._disk.get_folder_and_files(folder=source_folder, recursive=True)
        object_list = []
        for file in files:
            """
            'is_folder': this.is_folder,
            'relative_path': this.relative_path,
            'base64_content': this.base64_content,
            """
            an_object = {}
            an_object['relative_path'] = file.path
            if file.is_folder:
                an_object['is_folder'] = True
                an_object['base64_content'] = ""
            else:
                an_object['is_folder'] = False
                an_object['base64_content'] = self._disk.bytesio_to_base64(self._disk.get_bytesio_from_a_file(file.path)) 
            object_list.append(an_object)
        return json.dumps(object_list)

    def generate(self, source_folder: str, generated_file_path="lib/built_in_files.dart"):
        template1 = """
import 'dart:convert';
import 'dart:io';
import 'dart:typed_data';
import 'package:path/path.dart' as path_;
        """.strip()

        template2 = """
class File_Model {
  bool? is_folder;
  String? relative_path;
  String? base64_content;

  File_Model({this.is_folder, this.relative_path, this.base64_content});

  Map<String, dynamic> to_dict() {
    return {
      'is_folder': this.is_folder,
      'relative_path': this.relative_path,
      'base64_content': this.base64_content,
    };
  }

  File_Model from_dict(Map<String, dynamic>? json) {
    if (json == null) {
      return File_Model();
    }

    this.is_folder = json['is_folder'];
    this.relative_path = json['relative_path'];
    this.base64_content = json['base64_content'];

    return File_Model(
      is_folder: json['is_folder'],
      relative_path: json['relative_path'],
      base64_content: json['base64_content'],
    );
  }
}

Future<Uint8List> decode_base64_string(String content) async {
  return base64Decode(content.replaceAll(RegExp(r'\s'), ''));
}

Future<void> _write_base64_string_to_file(
    String content, String file_path) async {
  Uint8List bytes_data = await decode_base64_string(content);
  await File(file_path).parent.create(recursive: true);
  await File(file_path).writeAsBytes(bytes_data);
}

Future<void> _write_file_object_to_disk(
    File_Model a_file, String parent_folder) async {
  if (a_file.is_folder == null ||
      a_file.relative_path == null ||
      a_file.base64_content == null) {
    return;
  }

  String target_path = path_.join(parent_folder, a_file.relative_path!);

  if (a_file.is_folder == true) {
    if (!Directory(target_path).existsSync()) {
      await Directory(target_path).create(recursive: true);
    }
  } else {
    if (!File(target_path).existsSync()) {
      await _write_base64_string_to_file(
        a_file.base64_content!,
        target_path,
      );
    }
  }
}

Future<List<File_Model>> read_json_string_as_object_list() async {
  final data = await json.decode(the_json_data_that_honors_yingshaoxo);
  List<File_Model> files = [];
  for (final one in data) {
    files.add(File_Model().from_dict(one));
  }
  return files;
}

Future<void> release_all_built_in_files(String parent_folder_path) async {
  List<File_Model> files = await read_json_string_as_object_list();
  for (final one in files) {
    await _write_file_object_to_disk(one, parent_folder_path);
  }
}
        """.strip()

        content_string = self._get_content_json_string(source_folder=source_folder)

        middle_content = f'''
        String the_json_data_that_honors_yingshaoxo = """{content_string}""";
        '''.strip()

        self._io.write(file_path=generated_file_path, content=template1 + "\n\n" + middle_content + "\n\n" + template2)


if __name__ == "__main__":
    """
    from pprint import pprint
    disk = Disk()
    files = disk.get_files(os.path.abspath(".."))
    files = disk.sort_files_by_time(files)
    pprint(files)
    """
    """
    store = Store("test")
    store.set("o", "2")
    print(store.get_items())
    print(store.has_key("o"))
    print(store.get("ok", "alsjdkfla"))
    store.reset()
    print(store.get_items())
    """
    disk = Disk()
    files = disk.get_files(folder=".", type_limiter=[".mp4"])
    # print(disk.get_hash_of_a_path("/home/yingshaoxo/.python_history"))



from typing import Any


class Calculator():
    """
    A class for advanced mathematics
    """

    def __init__(self):
        import sympy
        self.sympy = sympy
        self.sympy.init_printing(use_unicode=True)

    def symbol(self, symbol_string):
        """
        define a unknown variable

        Parameters
        ----------
        symbol_string: string
            for example, `x`
        """
        return self.sympy.Symbol(symbol_string)

    def simplify(self, expression, use_latex=True):
        """
        simplify an expression

        Parameters
        ----------
        expression: string
            for example, `(x^2)^2`
        """
        r = self.sympy.simplify(expression)
        if use_latex == False:
            return str(r)
        else:
            r = self.sympy.latex(r)
            return r

    def differential(self, expression, variable, times=1, use_latex=True):
        """
        get derivative of an expression

        Parameters
        ----------
        expression: string
            for example, `(x^2)^2`

        variable: string or symbol
            for example, `x` or symbol('x')

        times: int
            if you want to get second derivative of an expression, you should set it to `2`
        """
        r = self.sympy.diff(expression, variable, times)
        if use_latex == False:
            return str(r)
        else:
            r = self.sympy.latex(r)
            return r

    def integral(self, expression, variable, use_latex=True):
        """
        get integration of an expression

        Parameters
        ----------
        expression: string
            for example, `(x^2)^2`

        variable: string or symbol
            for example, `x` or symbol('x')
        """
        r = self.sympy.integrate(expression, variable)
        if use_latex == False:
            return str(r)
        else:
            r = self.sympy.latex(r)
            return r


if __name__ == "__main__":
    pass



import pyfakewebcam
import cv2

from auto_everything.base import Terminal
t = Terminal(debug=True)


class FakeCamera():
    """
    This module based on `pyfakewebcam`.
    """

    def __init__(self, device: str = None, width: int = 640, height: int = 480):
        list_command = "v4l2-ctl --list-devices"
        list_devices = t.run_command(list_command)
        assert "not found" not in list_devices, f"{list_devices}\n\nv4l2-ctl not found, you may have to install it with 'sudo apt install v4l2loopback-utils'"
        assert "v4l2loopback" in list_devices, f"{list_devices}\n\nTo create a fake camera, you have to add a dummy device first by 'sudo modprobe v4l2loopback'"
        lines = list_devices.split("\n")
        the_device = None
        for index, line in enumerate(lines):
            if index < len(lines) - 1:
                if ("v4l2loopback" in line) and "/dev/video" in lines[index + 1]:
                    the_device = lines[index + 1].strip()
        if device == None:
            assert the_device != None, f"{list_devices}\n\ncan't auto find the dummy device, you may have to specify it manullay: FakeCamera('/dev/video*')"
            self.device = the_device
        else:
            self.device = device
        self.camera = pyfakewebcam.FakeWebcam(self.device, width, height)

    def next(self, frame, rgb=False):
        """
        Pipe new frame to the fake camera

        Parameters
        ----------
        frame: numpy representation of an image
            np.zeros((480,640,3), dtype=np.uint8)
        rgb: bool
            if you want to give me an rgb frame, you'll have to set this.
        """
        if rgb:
            frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
        self.camera.schedule_frame(frame)

    def show(self):
        """
        Show the fake camera by using ffplay

        Parameters
        ----------
        """
        t.run(f"ffplay {self.device}")


if __name__ == "__main__":
    from pprint import pprint
    fakecam = FakeCamera()



#sudo gcc main.cpp -lstdc++ -lX11 -lXmu -o main
from auto_everything.base import Terminal
t = Terminal()


def compile_python_module(cpp_name, extension_name):
    t.run(
        "g++ -O3 -Wall -Werror -shared -std=c++11 -fPIC "
        "`python3 -m pybind11 --includes` "
        "-I /usr/include/python3.7 -I .  "
        "{0} "
        "-o {1}`python3.7-config --extension-suffix` "
        "-L. -lX11 -lXmu -Wl,-rpath,.".format(cpp_name, extension_name)
    )


def build_cython():
    """ Build the cython extension module """
    print("* Start to build...")

    # Compile and link the cython wrapper library
    compile_python_module("myx11.cpp", "myx11")

    print("* Complete")


build_cython()



from typing import Any
import random
import os

from auto_everything.terminal import Terminal
from auto_everything.disk import Disk, Store
from auto_everything.io import IO
from auto_everything.language import Language
from auto_everything.time import Time
disk = Disk()
io_ = IO()
language = Language()
terminal = Terminal()
time_ = Time()
store = Store('auto_everything_ml_module')


class DataProcessor():
    """
    To implement some functionality related to data
    """

    def __init__(self):
        pass

    def get_time_series_data_from_a_list(self, the_list, sequence_length):
        """
        Get sub sequences for LSTM network.

        Parameters
        ----------
        the_list: 
        sequence_length: int
            how long you want the subsequence to be.

        Returns
        -------
        tuple
            return ([features], [labels])
        """
        assert len(the_list) >= sequence_length + 1, "len(the_list) should >= sequence_length + 1"
        array_1d = []
        array_2d = []
        array_target = []
        for element in the_list:
            array_1d.append(element)
            if len(array_1d) == sequence_length + 1:
                target = array_1d.pop()
                array_target.append(target)
                array_2d.append(array_1d.copy())
                array_1d.clear()
        return array_2d, array_target


class Yingshaoxo_Text_Generator():
    """
    1. First you have to have a folder where has multiple txt files
    2. This class will parse those text, convert it into 30000, 3000, 300, 30, 3, 1 char length sub_context_window, we slide that window by one char each search time
    3. If we found previous x_chars matchs the input_text user asks in our database, we return the following chars from database to the user
    4. Normally, we'll add a formater to the end of pipeline to format the final result to make it looks better.

    Second method is to do it with full-match, we only return full-match following text, if it can't get the following text, we do the search again with input_text[1:]
    And we could also use a transformer to get 1000 different ways format of the input_text, then do the full_match again.

    Third method:
        利用类似于谷歌search一样的东西
        加上问答系统，你也可以制作一个ChatGPT，并且准确率特别高
        举个例子，把input_text放进谷歌搜索，将第一页所有网页的内容作为问答系统的context
        准确率将高得惊人
        https://huggingface.co/distilbert-base-cased-distilled-squad?context=My+name+is+%E8%83%A1%E8%8B%B1%E6%9D%B0&question=What+is+my+name%3F

    Wha kind of problem I have solved using traditional programming?
    1. ChatBot
    2. Sentence translation
    3. Grammar correction
    4. Punctuation Correction Or Adding
    5. Code completion
    6. Sentence rewrite
    """
    def __init__(self, input_txt_folder_path: str, only_search_the_first_level_of_folders: bool = True, type_limiter: list[str] = [".txt", ".md"], use_machine_learning: bool = False, debug_mode: bool = False):
        self.debug_mode = debug_mode
        self.input_txt_folder_path = input_txt_folder_path

        self.text_source_data = ""
        files = disk.get_files(self.input_txt_folder_path, recursive=not only_search_the_first_level_of_folders, type_limiter=type_limiter)
        for file in files:
            self.text_source_data += io_.read(file)
            self.lower_case_text_source_data = self.text_source_data.lower()
        
        self.use_machine_learning = use_machine_learning
        if (use_machine_learning == True):
            # pip install sentence_transformers
            from sentence_transformers import SentenceTransformer, util
            self.sentence_transformers_model = SentenceTransformer('all-MiniLM-L6-v2')
            self.sentence_transformers_utility = util
    
    @staticmethod
    def get_random_text_deriation_from_source_text(source_text: str, random_remove_some_characters: bool = False, random_add_some_characters: bool = False, random_char_source_text: str = "") -> str:
        source_text_lines = source_text.split("\n")
        random.shuffle(source_text_lines)
        new_lines = []
        for line in source_text_lines:
            segments_list = language.seperate_text_to_segments(text=line, ignore_space=False)
            segments_list = [one["text"] for one in segments_list]
            random.shuffle(segments_list)
            new_line = "".join(segments_list)
            new_lines.append(new_line)
        final_random_text = "\n".join(new_lines)

        random_length = int(len(source_text) * 0.2)
        if random_remove_some_characters:
            for i in range(random_length):
                random_index = random.randint(0, len(final_random_text)-1)	
                final_random_text = final_random_text[:random_index] + final_random_text[random_index + 1:]
        if random_add_some_characters:
            for i in range(random_length):
                random_index = random.randint(0, len(final_random_text)-1)	
                if (random_char_source_text == ""):
                    random_char_source_text = source_text
                final_random_text = final_random_text[:random_index] + random.choice(random_char_source_text) + final_random_text[random_index:]

        return final_random_text
    
    def get_similarity_of_two_sentences(self, sentence_1: str, sentence_2: str, use_both_machine_learning_and_traditional_method: bool = False) -> float:
        if use_both_machine_learning_and_traditional_method == True:
            sentence_embedding_list = self.sentence_transformers_model.encode(sentences=[sentence_1, sentence_2], convert_to_tensor=True)
            similarity = self.sentence_transformers_utility.cos_sim(sentence_embedding_list[0], sentence_embedding_list[1])
            similarity1 = float(similarity.cpu().numpy()[0][0])
            similarity2 = language.compare_two_sentences(sentence_1, sentence_2)
            return (similarity1 + similarity2) / 2
        else:
            if self.use_machine_learning == True:
                sentence_embedding_list = self.sentence_transformers_model.encode(sentences=[sentence_1, sentence_2], convert_to_tensor=True)
                similarity = self.sentence_transformers_utility.cos_sim(sentence_embedding_list[0], sentence_embedding_list[1])
                return float(similarity.cpu().numpy()[0][0])
            else:
                return language.compare_two_sentences(sentence_1, sentence_2)
    
    def _count_how_many_sub_string_in_previous_context(self, start_index: int, input_text: str, how_long_the_text_you_want_to_get: int = 1024):
        input_text = input_text.lower()

        all_substring_list = []
        for index, _ in enumerate(input_text):
            for index2, _ in enumerate(input_text[index:]):
                index2 = index + index2 + 1
                sub_string = input_text[index: index2]
                all_substring_list.append(sub_string)
        all_substring_list.sort(key=len, reverse=True)
        all_substring_list = all_substring_list[:len(all_substring_list)//2]

        new_source_text = self.lower_case_text_source_data[start_index-how_long_the_text_you_want_to_get: start_index]
        counting = 0
        for index, sub_string in enumerate(all_substring_list):
            if sub_string in new_source_text:
                counting += len(sub_string)
        return counting

    def search_and_get_following_text(self, input_text: str, quick_mode: bool = True, use_fuzz_search: bool = True, how_long_the_text_you_want_to_get: int = 1024) -> tuple[str, str]:
        """
        It will return you the context and following text as a format of tuple[context, following_text]
        """
        if (input_text.strip() == ""):
            return "", ""

        input_text = input_text.lower()

        found_dict = {}
        search_start_index = 0
        while True:
            found = self.lower_case_text_source_data.find(input_text, search_start_index)
            if found == -1:
                # didn't found
                break
            else:
                start = found
                end = found + len(input_text)
                found_dict[found] = {
                    "start": start,
                    "end": end,
                    "following": self.text_source_data[end: end + how_long_the_text_you_want_to_get]
                }
                search_start_index = start + 1

                #if quick_mode == True:
                #    break

        if len(found_dict.keys()) > 0:
            random_key = random.choice(list(found_dict.keys())) 
            return self.text_source_data[found_dict[random_key]["end"]-how_long_the_text_you_want_to_get:found_dict[random_key]["end"]+how_long_the_text_you_want_to_get], found_dict[random_key]["following"]
        else:
            if use_fuzz_search == False:
                return self.search_and_get_following_text(input_text = input_text[len(input_text)//2+1:], quick_mode = True, use_fuzz_search = True, how_long_the_text_you_want_to_get = how_long_the_text_you_want_to_get)
            else:
                if (self.debug_mode):
                    print("Using fuzz searching...")

                all_substring_list = []
                for index, _ in enumerate(input_text):
                    for index2, _ in enumerate(input_text[index:]):
                        index2 = index + index2 + 1
                        sub_string = input_text[index: index2]
                        all_substring_list.append(sub_string)
                all_substring_list.sort(key=len, reverse=True)
                all_substring_list = all_substring_list[:len(all_substring_list)//2]

                # what I did here is simply try to search keywords(sub_strings) in previous text, the more matchs, that part of text if more likely the one we are looking for.
                # It can get improved by using some word_spliting library, and even more, you can add same_meaning_words library to it to make sure it always finds out the right data.
                possibility_list = []
                for sub_string in all_substring_list:
                    search_start_index = 0
                    highest_counting = 0
                    highest_counting_info_dict = None
                    while True:
                        found = self.lower_case_text_source_data.find(sub_string, search_start_index)
                        if found == -1:
                            # didn't found
                            break
                        else:
                            start = found
                            end = found + len(input_text)
                            info_dict = {
                                "start": start,
                                "end": end,
                                "following": self.text_source_data[end: end + how_long_the_text_you_want_to_get],
                                "relative_counting": 0
                            }
                            search_start_index = start + 1

                            relative_counting = self._count_how_many_sub_string_in_previous_context(start_index=end, input_text=sub_string, how_long_the_text_you_want_to_get=how_long_the_text_you_want_to_get)
                            if relative_counting > highest_counting:
                                highest_counting = relative_counting
                                info_dict["relative_counting"] = relative_counting
                                highest_counting_info_dict = info_dict.copy()

                                if quick_mode == True:
                                    break

                    if highest_counting_info_dict != None:
                        possibility_list.append(highest_counting_info_dict.copy())

                if len(possibility_list) > 0:
                    possibility_list.sort(key=lambda item: item["relative_counting"], reverse=True)
                    return self.text_source_data[possibility_list[0]['end']-how_long_the_text_you_want_to_get:possibility_list[0]['end']+how_long_the_text_you_want_to_get], possibility_list[0]["following"]
                else:
                    return self.search_and_get_following_text(input_text = input_text[len(input_text)//2+1:], quick_mode = quick_mode, use_fuzz_search = use_fuzz_search, how_long_the_text_you_want_to_get = how_long_the_text_you_want_to_get)

    def search_and_get_following_text_in_a_exact_way(self, input_text: str, quick_mode: bool = False, use_fuzz_search: bool = True, extremly_accrate_mode: bool = False, how_long_the_text_you_want_to_get: int = 1024, also_want_the_current_line: bool = False) -> str:
        context, following_text = self.search_and_get_following_text(input_text=input_text, quick_mode=quick_mode, use_fuzz_search=use_fuzz_search, how_long_the_text_you_want_to_get=how_long_the_text_you_want_to_get)
        if (context.strip() == ""):
            return "..."

        context_splits = language.seperate_text_to_segments(context)
        input_text_splits = language.seperate_text_to_segments(input_text)
        # context_splits = context.split("\n") 
        # input_text_splits = [one for one in input_text.split("\n") if one.strip() != ""]

        last_input_sentence = ""
        for one_input in reversed(input_text_splits):
            if one_input["is_punctuation_or_space"] == False:
                last_input_sentence = one_input["text"]
                if (self.debug_mode):
                    print(f"last_input_sentence: {last_input_sentence}")
                break
        
        similarity_list = []
        for index, one_target in enumerate(context_splits):
            if one_target ["is_punctuation_or_space"] == False:
                one_sentence = one_target["text"]
                if (self.use_machine_learning):
                    similarity = self.get_similarity_of_two_sentences(one_sentence, last_input_sentence)
                else:
                    similarity = language.compare_two_sentences(one_sentence, last_input_sentence)
                similarity_list.append({
                    "similarity": similarity,
                    "start_index": index
                })
        similarity_list.sort(key=lambda item: item["similarity"], reverse=True)
        the_seperator_index = similarity_list[0]["start_index"]

        if extremly_accrate_mode == True:
            for index, one_target in enumerate(context_splits):
                if one_target ["is_punctuation_or_space"] == False:
                    one_sentence = one_target["text"]
                    if one_sentence.lower() == last_input_sentence.lower():
                        the_seperator_index = index
                        break

        if (also_want_the_current_line == False):
            for index, one in enumerate(context_splits[the_seperator_index:]):
                if one["is_punctuation_or_space"] == False:
                    the_seperator_index += index
                    break
        else:
            the_seperator_index -= 1

        return "".join([one["text"] for one in context_splits[the_seperator_index:]])
    
    @staticmethod
    def next_code_generation(data_source_folder_path: str, input_text: str, quck_mode: bool = True, type_limiter: list[str] = [".txt", ".py", ".md"], how_long_the_text_you_want_to_get: int = 1024):
        """
        1. take the previous text as input
        2. take sub_string of the input_text, from right to left, from long to short.
        3. search the database source text, if that sub_string matchs, add len(sub_string) to variable {one_following_char: count + len(sub_string)}
        4. take the biggest counting char as the next char

        method1: previous_text[-i:], search from i == 0 to i == len(previous_text), until it founds nothing, then go back, choose a random one
        method2: previous_text[i:], search from i == 0 to i == len(previous_text), for each time, i=i*2, until it found something, return that
        """
        text_source_data = ""
        should_update_datasource = False

        datestamp_string = store.get('last_code_generation_database_update_time', None)
        if (datestamp_string == None):
            should_update_datasource = True
        else:
            old_time = time_.get_datetime_object_from_timestamp(int(datestamp_string))
            new_time = time_.get_datetime_object_from_timestamp(time_.get_current_timestamp_in_10_digits_format())
            if (new_time - old_time).days > 3: #update the database for every 3 days
                should_update_datasource = True

        data_source_txt_file_path = terminal.fix_path("~/.auto_everything/ml/code_completion_data_source.txt")
        disk.create_a_folder(disk.get_directory_path(data_source_txt_file_path))
        if (not disk.exists(data_source_txt_file_path)):
            io_.write(file_path=data_source_txt_file_path, content="")

        if should_update_datasource == True:
            files = disk.get_files(folder=terminal.fix_path(data_source_folder_path), type_limiter=type_limiter)
            io_.write(file_path=data_source_txt_file_path, content="")
            for file in files:
                io_.append(file_path=data_source_txt_file_path, content=io_.read(file) + "\n\n\n\n")
            store.set('last_code_generation_database_update_time', str(time_.get_current_timestamp_in_10_digits_format()))
        else:
            text_source_data = io_.read(data_source_txt_file_path)
        
        def real_next_code_generation(input_text: str, how_long_the_text_you_want_to_get: int = 1024):
            if (input_text.strip() == ""):
                return ""

            found_start_index = text_source_data.find(input_text)
            if found_start_index == -1:
                # didn't found
                if quck_mode == True:
                    input_text = input_text[len(input_text)//2+1:]
                else:
                    input_text = input_text[1:]
                return real_next_code_generation(input_text=input_text, how_long_the_text_you_want_to_get=how_long_the_text_you_want_to_get)
            else:
                start = found_start_index
                end = found_start_index + len(input_text)
                following = text_source_data[end: end + how_long_the_text_you_want_to_get]
                following = following.rstrip()
                return following
        
        return real_next_code_generation(input_text=input_text, how_long_the_text_you_want_to_get=how_long_the_text_you_want_to_get)

    def text_to_text_harding_coding_transforming(self, input_text_list: list[str], output_text_list: list[str], text_you_want_to_transform: str):
        """
        1. Just think the whole transforming process as doing the search in a Q table.
        2. You use a patten filter to check the input_text, "I love you", 3 elements as a window, then you use this patten to do a search in the Q table, you found ["I hate you", "I trust you", "I hate you"], it seems like 'hate' has higher chance to be in the middle of that sentence.
        3. Or, you can simply think this: For a list of "I * you" patten in dataset, what word has more frequency in the position of *?, Choose the one has higher frequency.
        4. tip 3 is still in [MASK] level. If you want to handle the sentence segment sorting problem, you have to predict the 'move farwrd x characters' and 'move backword x chracter' information. Which can also be treated like a mask.

        speak of the process speed, use cache.
        
        this function could be used on 'wrong word correction', 'punctuation adding', 'sub_sentence rewrite'

        for 'summarytion task', get substrings from source_text, then get substrings from target_text, see how many substring shold get removed, get couting of those substrings that should get removed over the whole dataset.
            for the next time, in a new input sentence, we get those substring ranks, simply remove those substring that has higher 'garbage rank number'
        
        for 'sorting task', get substrings from input_text, then try to use before_context and following_context to do a search in target dataset, get the percentage of start_index/the_whole_length_of_the_sentence. 
            do a compare for the substring in the input_text, so you would get a percentage number of weather to move that substring farward or backward.
        """
        print("Haven't get implemented yet.")
        # def _count_how_many_sub_string_in_previous_context(self, start_index: int, input_text: str, how_long_the_text_you_want_to_get: int = 1024):
        #     input_text = input_text.lower()

        #     all_substring_list = []
        #     for index, _ in enumerate(input_text):
        #         for index2, _ in enumerate(input_text[index:]):
        #             index2 = index + index2 + 1
        #             sub_string = input_text[index: index2]
        #             all_substring_list.append(sub_string)
        #     all_substring_list.sort(key=len, reverse=True)
        #     all_substring_list = all_substring_list[:len(all_substring_list)//2]

        #     new_source_text = self.lower_case_text_source_data[start_index-how_long_the_text_you_want_to_get: start_index]
        #     counting = 0
        #     for index, sub_string in enumerate(all_substring_list):
        #         if sub_string in new_source_text:
        #             counting += len(sub_string)
        #     return counting


class Yingshaoxo_Computer_Vision():
    def __init__(self):
        import numpy
        self.numpy = numpy

    def get_similarity_of_two_images(self, numpy_image_1: Any, numpy_image_2: Any) -> float:
        """
        return a float between 0 and 1, 1 means equal, 0 means no relate.
        """
        mean1 = self.numpy.mean(numpy_image_1, axis=(0, 1))
        mean2 = self.numpy.mean(numpy_image_2, axis=(0, 1))

        difference = 0.0
        difference += self.numpy.absolute(mean1[0] - mean2[0])
        difference += self.numpy.absolute(mean1[1] - mean2[1])
        difference += self.numpy.absolute(mean1[2] - mean2[2])

        final_difference = ((difference * 100) / (255*3)) / 20
        final_difference = 1 - final_difference
        return final_difference


class Yingshaoxo_Speech_Recognizer():
    def __init__(self, language: str = 'en'):
        # pip install vosk
        # pip install sounddevice
        import queue
        import sys
        import sounddevice
        from vosk import Model, KaldiRecognizer
        from auto_everything.time import Time

        self.queue = queue
        self.sys = sys
        self.sounddevice = sounddevice
        self.time_ = Time()

        if language == "en":
            self.vosk_model = Model(lang="en-us")
        else:
            self.vosk_model = Model(model_name="vosk-model-cn-0.22")

        self.KaldiRecognizer = KaldiRecognizer

        self.microphone_bytes_data_queue = queue.Queue()
    
    def recognize_following_speech(self, timeout_in_seconds: int | None = None) -> str:
        while self.microphone_bytes_data_queue.empty() == False:
            self.microphone_bytes_data_queue.get_nowait()

        def callback(indata, frames, time, status):
            """This is called (from a separate thread) for each audio block."""
            if status:
                print(status, file=self.sys.stderr)
            self.microphone_bytes_data_queue.put(bytes(indata))

        try:
            device_info = self.sounddevice.query_devices(None, "input")
            samplerate = int(device_info["default_samplerate"]) #type:ignore
                
            with self.sounddevice.RawInputStream(samplerate=samplerate, blocksize = 8000, device=None,
                    dtype="int16", channels=1, callback=callback):
                rec = self.KaldiRecognizer(self.vosk_model, samplerate)

                start_time = self.time_.get_current_timestamp_in_10_digits_format()
                while True:
                    data = self.microphone_bytes_data_queue.get()
                    if rec.AcceptWaveform(data):
                        text = json.loads(rec.Result())["text"] #type:ignore
                        text = text.replace(" ", "").strip()
                        if len(text) != 0:
                            #print(text)
                            return text
                    else:
                        # print(rec.PartialResult())
                        pass
                    end_time = self.time_.get_current_timestamp_in_10_digits_format()
                    if timeout_in_seconds != None:
                        duration = self.time_.get_datetime_object_from_timestamp(end_time) - self.time_.get_datetime_object_from_timestamp(start_time)
                        if duration.seconds > timeout_in_seconds:
                            return ""
        except Exception as e:
            print(e)
            return ""


class Yingshaoxo_Translator():
    def __init__(self):
        # pip install dl-translate
        import dl_translate
        from auto_everything.language import Language
        self.dl_translate = dl_translate
        self.dl_translate_model = self.dl_translate.TranslationModel(device="auto")
        self.languages = self.dl_translate.lang
        self._language = Language()
    
    def translate(self, text: str, from_language: Any, to_language: Any, sentence_seperation: bool = False) -> str:
        try:
            text = text.strip()
            if sentence_seperation == True:
                data_list = self._language.seperate_text_to_segments(text=text, ignore_space=True)
                """
                [
                    {
                        "is_punctuation_or_space": true, "text": "?",
                    }, {
                        "is_punctuation_or_space": false, "text": "Yes",
                    },
                ]
                """
                text_list = []
                for segment in data_list:
                    if segment["is_punctuation_or_space"] == False:
                        result = self.dl_translate_model.translate(segment["text"], source=from_language, target=to_language)
                        result = str(result).strip("!\"#$%&'()*+, -./:;<=>?@[\\]^_`{|}~ \n，。！？；：（）［］【】")
                        text_list.append(result)
                    else:
                        text_list.append(segment["text"])
                return "".join(text_list)
            else:
                return self.dl_translate_model.translate(text, source=from_language, target=to_language) #type: ignore
        except Exception as e:
            print(e)
            return text
    
    def chinese_to_english(self, text: str, sentence_seperation: bool = False):
        return self.translate(text=text, from_language=self.languages.CHINESE, to_language=self.languages.ENGLISH, sentence_seperation=sentence_seperation)

    def english_to_chinese(self, text: str, sentence_seperation: bool = False):
        return self.translate(text=text, from_language=self.languages.ENGLISH, to_language=self.languages.CHINESE, sentence_seperation=sentence_seperation)


class Yingshaoxo_Text_to_Speech():
    def __init__(self):
        #pip install TTS
        #sudo apt install ffmpeg                 or          https://github.com/markus-perl/ffmpeg-build-script#:~:text=maintain%20different%20systems.-,Installation,-Quick%20install%20and
        os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"

        from TTS.api import TTS
        self.TTS = TTS

        from auto_everything.terminal import Terminal
        from auto_everything.disk import Disk
        self.terminal = Terminal()
        self.disk = Disk()

        import torch
        use_gpu = True if torch.cuda.is_available() else False
        self.torch = torch
        #pprint(TTS.list_models())

        self.tts_en = TTS("tts_models/en/ljspeech/tacotron2-DDC", gpu=use_gpu)
        # self.tts_en = TTS("tts_models/en/ljspeech/fast_pitch", gpu=use_gpu)
        self.tts_cn = TTS("tts_models/zh-CN/baker/tacotron2-DDC-GST", gpu=use_gpu)

    def _language_splitor(self, text: str):
        language_list = []
        index = 0
        while True:
            temp_string = ""
            if (index >= len(text)):
                break
            char = text[index]
            while ord(char) < 128:
                # english
                char = text[index]
                temp_string += char
                index += 1
                if (index >= len(text)):
                    break
            if (temp_string.strip() != ""):
                temp_string = temp_string[:-1]
                index -= 1
                language_list.append({
                    "language": "en",
                    "text": temp_string
                })

            temp_string = ""
            if (index >= len(text)):
                break
            char = text[index]
            while not ord(char) < 128:
                # chinese 
                char = text[index]
                temp_string += char
                index += 1
                if (index >= len(text)):
                    break
            if (temp_string.strip() != ""):
                temp_string = temp_string[:-1]
                index -= 1
                language_list.append({
                    "language": "cn",
                    "text": temp_string
                })

            if (index+1 >= len(text)):
                break

        if len(language_list) > 0:
            language_list[-1]["text"] += text[-1]
        
        new_list = []
        for index, one in enumerate(language_list):
            new_text = language_list[index]["text"].strip()
            if len(new_text) > 0:
                new_list.append({
                    'language': one['language'],
                    'text': new_text
                })

        return new_list

    def _speak_it(self, language: str, text: str):
        output_file = os.path.abspath(os.path.join(self.disk.get_a_temp_folder_path(), "output.wav"))
        self.disk.create_a_folder(self.disk.get_directory_path(output_file))

        text = text.strip("!\"#$%&'()*+, -./:;<=>?@[\\]^_`{|}~ \n，。！？；：（）［］【】")
        if (language == "en"):
            tts = self.tts_en
            text += "."
        else:
            tts = self.tts_cn
            text += "。"

        try:
            if tts.speakers == None:
                tts.tts_to_file(text=text, file_path=output_file)
            else:
                tts.tts_to_file(text=text, file_path=output_file, speaker=tts.speakers[0], language=tts.languages[0], speed=2.5) #type:ignore
        except Exception as e:
            print(e)

        self.terminal.run(f"""
        ffplay -autoexit -nodisp "{output_file}"
                """, wait=True)
        
        self.disk.delete_a_file(output_file)

    def speak_it(self, text: str):
        data_ = self._language_splitor(text)
        for one in data_:
            print(one)
            self._speak_it(language=one["language"], text=one["text"])


if __name__ == "__main__":
    pass
